/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "base.h"
#include "log.h"

namespace SpritePack
{

	Base::Base()
		: jpegQuality(89),
		pngCompressionLevel(6),
		pngFilter(-1),
		startupMs(100),
		allowLossyTypeChange(true),
		k("10%"),
		usedK(1),
		classesCount(0),
		inputCost(0),
		inputFileSize(0),
		inputArea(0),
		loadFinishedTime(0),
		useBL(true),
		useMBL(false),
		useVHTL(false),
		useTwoFit(false),
		useFFDH(false),
		useNFDH(false),
		useBFDH(false),
		nullImage(Image()),
		nullImageInfo()
	{
		speedMbpsTable.push_back(0);
	}

	Base::~Base()
	{
		closeLogFile();
	}

	bool Base::isValid(Id id) const
	{
		return 0 <= id && id < (int)images.size();
	}

	const ImageInfo& Base::imageInfo(Id id) const
	{
		return isValid(id) ? imageInfos[id] : nullImageInfo;
	}

	const Image& Base::image(Id id) const
	{
		return isValid(id) ? images[id] : nullImage;
	}

	Base& Base::operator=(const Base&)
	{
		return *this;
	}

	double Base::speedMbps(int c) const
	{
		if (c <= 0 || speedMbpsTable.size() <= 1)
			return 0;
		
		if (c >= speedMbpsTable.size())
		{
			int lastC = (int)speedMbpsTable.size() - 1;
			return speedMbpsTable[lastC] * lastC / c;
		}

		return speedMbpsTable[c];
	}

	Cost Base::inifnity()
	{
		return Cost::infinity();
	}

	string Base::generateCssClassName(const string& fileName, int i) const
	{
		string filePathLower = toLower(pathNormalize(fileName));
		string ext;
		size_t extPos;

		extPos = filePathLower.find_last_of('.');
		if (extPos != string::npos)
		{
			ext = filePathLower.substr(extPos + 1);
			filePathLower = filePathLower.substr(0, extPos);
		}

		filePathLower = ltrim(filePathLower, "0123456789");

		for (char & c : filePathLower)
		{
			if ((c < '0' || '9' < c) && (c < 'a' || 'z' < c) && c != '_')
				c = '-';
		}

		filePathLower = replaceAll(filePathLower, "--", "-", true);

		filePathLower = trim(filePathLower, "-");

		if (i == 0)
		{
			return "img-" + filePathLower;
		}
		else 
		{
			return "img-" + filePathLower + toString(i);
		}
	}

	bool Base::isGoodExt(const string& fileName) const
	{
		return Type(fileExt(fileName)) != None;
	}

	vector<Palette> Base::palettes() const
	{
		const Palette nullPalette = Palette();
		set<Palette> s;
		for (const ImageInfo& info : imageInfos)
			if (info.palette != nullPalette)
				s.insert(info.palette);

		return vector<Palette>(s.begin(), s.end());
	}

	void Base::setClassStats(const string& name, int n)
	{
		if (n <= 0)
			return;

		map<string, int>::iterator itr = classesStats.find(name);
		if (itr == classesStats.end())		
			classesStats.insert(make_pair(name, n));
		else		
			itr->second += n;	
	}

	void Base::compressionMoveUsed(const string& name)
	{
		if (name.empty())
			return;

		map<string, int>::iterator itr = compressionMovesStats.find(name);
		if (itr == compressionMovesStats.end())		
			compressionMovesStats.insert(make_pair(name, 1));
		else		
			++itr->second;		
	}

	void Base::geometryMoveUsed(const string& name)
	{
		if (name.empty())
			return;

		map<string, int>::iterator itr = geometryMovesStats.find(name);
		if (itr == geometryMovesStats.end())		
			geometryMovesStats.insert(make_pair(name, 1));
		else		
			++itr->second;		
	}

	vector<int> Base::pngFilters() const
	{
		vector<int> filters;
		if (1 <= pngFilter && pngFilter < 10)
		{
			filters.push_back(pngFilter);
		}
		else
		{
			for (int i = 0; i < 10; ++i)
				filters.push_back(i);
		}
		return filters;
	}

	void Base::calcBestTypePalette(const string& fileName, Image * input, Type* bestType, Palette* bestPalette, FileSize* bestFileSize, Quality* bestQuality) const
	{
		if (!input || !bestType || !bestPalette || !bestFileSize || !bestQuality)
			return;

		string tmpFileName;

		const string cmdBegins[] = { "identify -ping -verbose", "/usr/bin/identify -ping -verbose" };

		string bestFileName = fileName;

		if (Type(input->magick()) != None)
		{
			Magick::Blob bestBlob;
			{
				Image& img = *input; //for lossy save, consider copy
				*bestType = Type(img.magick());
				*bestFileSize = img.fileSize();
				if (typeHasQuality(*bestType))
					*bestQuality = (int)img.quality();

				for (int filter : pngFilters())
				{
					img.quality(pngCompressionLevel * 10 + filter % 10);
					img.type(Magick::OptimizeType);
					img.strip();
					img.magick(Type(Png).toString());

					Magick::Blob blob;
					img.write(&blob);
					if ((FileSize)blob.length() < *bestFileSize)
					{
						bestBlob = blob;
						*bestType = Type(img.magick());
						*bestFileSize = blob.length();
						*bestQuality = (int)img.quality();
					}

				}
			}

			if (bestBlob.length() > 0)
			{
				tmpFileName = createTmpFilePath();
				bestFileName = tmpFileName;
				size_t len = bestBlob.length();
				fstream f(tmpFileName.c_str(), fstream::out | fstream::trunc | fstream::binary);
				f.write((const char*)bestBlob.data(), len);							
			}
			

			bool readPaletteFromFile = true;

			if (*bestType == Jpeg)
			{
				*bestPalette = Palette("Jpeg", false);
				readPaletteFromFile = false;
			}

			if (readPaletteFromFile)
			{
				const string stdOutFileName = createTmpFilePath();
				string stdOutData;
				string cmd;
				int ret = 0;
				for (string cmdBegin: cmdBegins)
				{
					const string bestFileName2 = "\"" + replaceAll(bestFileName, "\"", "\\\"") + "\"";
					const string stdOutFileName2 = "\"" + replaceAll(stdOutFileName, "\"", "\\\"") + "\"";
					cmd = cmdBegin + " " + bestFileName2 + " > " + stdOutFileName2;
					ret = system(cmd.c_str());
					if (ret == 0)
						break;
				}
				if (ret == 0)
				{
					ifstream f(stdOutFileName.c_str(), fstream::binary);
					if (f.is_open())
					{
						f.seekg(0, ios::end);
						size_t length = f.tellg();
						f.seekg(0, ios::beg);
						if (length > 0) 
						{
							vector<char> data(length, 0);
							f.read(data.data(), length);
							stdOutData = string(data.data());
						}
					}
				}

				if (!stdOutFileName.empty())
				{
					removeFile(stdOutFileName);	
				}

				if (!tmpFileName.empty())
				{
					removeFile(tmpFileName);
					tmpFileName = "";
				}

				if (ret != 0 || stdOutData.empty())
				{
					throwError("Cannot execute: " + cmd);
				}			

				const string imgType = readStringValue(stdOutData, "Type:");
				const int pngDepth = readIntValue(stdOutData, "png:IHDR.bit_depth");
				const string pngColorType = readStringValue(stdOutData, "png:IHDR.color_type");
				const int redDepth = readIntValue(stdOutData, "red:");
				const int greenDepth = readIntValue(stdOutData, "green:");
				const int blueDepth = readIntValue(stdOutData, "blue:");
				const int alphaDepth = readIntValue(stdOutData, "alpha:");

				const bool isIndexed = toLower(imgType).find("palette") != string::npos;
				const bool isGrayscale = toLower(pngColorType).find("grayscale") != string::npos;

				const bool hasTransparency = alphaDepth != 0;

				int depth = redDepth + greenDepth + blueDepth + alphaDepth;

				if (pngDepth > 0)
					depth = min(depth, pngDepth);

				if (isIndexed)
					*bestPalette = Palette("Indexed", hasTransparency);
				else if (isGrayscale)
					*bestPalette = Palette("Grayscale", hasTransparency);
				else
					*bestPalette = Palette("TrueColor", hasTransparency);

			}

		}


		if (!tmpFileName.empty())
		{
			removeFile(tmpFileName);
			tmpFileName = "";
		}
	}

	void Base::readK(int * numberValue, double * percentValue) const
	{
		if (!numberValue || !percentValue)
			return;

		*numberValue = 0;
		*percentValue = 0;

		size_t percentPos = k.find('%');
		const bool isPercent = percentPos != string::npos;
		if (isPercent)
		{
			string s = k.substr(0, percentPos);
			double percent = fromString<double>(s);
			if (0 < percent && percent <= 100)
			{
				*percentValue = percent;
			}
		}
		else
		{
			*numberValue = fromString<int>(k);
		}
	}

	void Base::checkInputParams() const
	{
		int numberK = 0;
		double percentK = 0;
		readK(&numberK, &percentK);

		if (numberK == 0 && percentK == 0)
			throwError("Wrong k: " + k);

		if (jpegQuality <= 0 || jpegQuality > 100)
			throwError("Wrong jpegQuality: " + toString(jpegQuality));

		if (pngCompressionLevel < 0 || pngCompressionLevel > 9)
			throwError("Wrong pngCompressionLevel: " + toString(pngCompressionLevel));

		if (pngFilter != -1 && (pngFilter < 0 || pngFilter > 9))
			throwError("Wrong pngFilter: " + toString(pngFilter));

		if (!useBL && !useMBL && !useFFDH && !useBFDH && !useNFDH && !useVHTL && !useTwoFit)
			throwError("No algorithms will be used");

		if (speedMbpsTable.size() <= 1)
			throwError("No speed table");

		/*
		if (startupMs < 0)
			throwError("Wrong startupMs: " + toString(startupMs));

		if (speedMbps < 0)
			throwError("Wrong speedMbps: " + toString(speedMbps));
		*/
	}

	void Base::checkInputDir() const
	{
		bool ok = false;
		try
		{
			ok = !readdir(inputDir).empty();
		}
		catch (...)
		{
		}

		if (!ok)
			throwError("Empty or wrong inputDir: " + inputDir);		
	}

	void Base::checkOutputDir() const
	{
		bool ok = false;
		try
		{
			readdir(outputDir);
			ok = true;
		}
		catch (...)
		{
		}
		if (!ok)
			throwError("Wrong outputDir: " + outputDir);
		
	}

	void Base::openLogFile(const string& _fileName)
	{
		string fileName = _fileName;
		if (fileName.empty()) 
			fileName = pathJoin(outputDir, "log.txt");
		Log::openLogFile(fileName);
	}

	void Base::closeLogFile()
	{
		Log::closeLogFile();
	}
	
	pair<ImageInfo, Image> Base::loadImage(const string& fileName) const
	{
		const string fullFileName = pathJoin(inputDir, fileName);
		list<Image> imageList;
		if (isGoodExt(fullFileName))
		{
			try
			{
				readImages(&imageList, fullFileName);
			}
			catch (...)
			{
				imageList.clear();
			}
		}
		if (!imageList.empty())
		{
			Image img(imageList.front());
			Type type(img.magick());
			if (type != None && imageList.size() == 1)
			{
				imageList.clear();
				const int width = (int)img.columns();
				const int height = (int)img.rows();
				Area area = (Area)width * (Area)height;
				if (area > 0)
				{
					Palette palette;
					FileSize beforeFileSize = img.fileSize();
					FileSize afterFileSize = img.fileSize();
					Quality quality = (int)img.quality();
					calcBestTypePalette(fullFileName, &img, &type, &palette, &afterFileSize, &quality);

					return pair<ImageInfo, Image>(ImageInfo(-1, fileName, "", width, height, type, quality, palette, beforeFileSize, afterFileSize), img);
				}
			}
		}

		return pair<ImageInfo, Image>();
	}

	void Base::load()
	{
		::srand(::time(NULL));
		showInputParamsInfo();

		images.clear();
		imageInfos.clear();
		inputCost = 0;
		inputFileSize = 0;
		inputArea = 0;

		vector<Cost> inputCosts;
		set<string> cssClassesSet;
		list<string> listing = readdir(inputDir);
		vector<string> vectorListing(listing.begin(), listing.end());
		sort(vectorListing.begin(), vectorListing.end());

		Log() << "Started";
		for (string fileName : vectorListing)
		{
			pair<ImageInfo, Image> p = loadImage(fileName);
			ImageInfo info = p.first;
			if (info.width > 0)
			{
				inputCosts.push_back(costFromFileSize(info.beforeFileSize));

				inputFileSize += info.beforeFileSize;
				inputArea += (Area)info.width * (Area)info.height;

				string cssClassName;
				for (int cssI = 0; ;++cssI)
				{
					cssClassName = generateCssClassName(fileName, cssI);
					if (cssClassesSet.find(cssClassName) == cssClassesSet.end())
						break;
				}
				cssClassesSet.insert(cssClassName);

				images.push_back(p.second);
				imageInfos.push_back(ImageInfo((int)(images.size() - 1), fileName, cssClassName, info.width, info.height, info.type, info.quality, info.palette, info.beforeFileSize, info.afterFileSize));

				Log() << "File:" << fileName << info.type.toString() << info.palette.toString() << toString(info.palette.hasTransparency()) << -(info.afterFileSize - info.beforeFileSize);

			}
			else
				Log() << "Not supported file:" << fileName;
			
		}
		loadFinishedTime = Time::elapsed();
		inputCost = costFunction(inputCosts);
		Log() << "Loaded, cost: " + inputCost.toString(CompressionCost) + ", filesize: " + toString(inputFileSize) + ", count: " + toString(imageInfos.size()) + ", area: " + toString(inputArea);
	}

	bool Base::typeHasQuality(const Type& type) const
	{
		switch (type)
		{
		case Png:
		case Jpeg:
			return true;
		default:
			return false;
		}
	}

	Geometry Base::geometry(Id id) const
	{
		return imageInfo(id).geometry();
	}

	Geometry Base::geometry(const Bucket::Items& items) const
	{
		pair<int, int> result(0, 0);
		for (const Bucket::Item& item : items)
		{
			if (isValid(item))
			{
				const ImageInfo& info = imageInfo(item);
				result.first = max<int>(result.first, item.x + info.width);
				result.second = max<int>(result.second, item.y + info.height);
			}
		}
		return Geometry(result.first, result.second);
	}

	Geometry Base::geometry(const Bucket& bucket) const
	{
		return geometry(bucket.items());
	}

	Color Base::background(const Bucket& bucket) const
	{
		return bucket.type() == Png ? Color(0, 0, 0, 0) : Magick::ColorRGB(1.0, 1.0, 1.0);
	}

	bool Base::isLossType(Type type) const
	{
		switch (type)
		{
		case Jpeg:
			return true;
		case Gif:
		case Png:
			return false;
		default:
			return false;
		}
	}

	bool Base::hasTransparency(const Bucket& bucket) const
	{
		for(const Bucket::Item& item : bucket.items())
			if (imageInfo(item.id).palette.hasTransparency())
				return true;

		return false;
	}

	bool Base::hasType(const Bucket& bucket, Type type) const
	{
		for(const Bucket::Item& item : bucket.items())
			if (imageInfo(item.id).type == type)
				return true;

		return false;
	}

	bool Base::hasLossless(const Bucket& bucket) const
	{
		for(const Bucket::Item& item : bucket.items())
			if (isLossType(imageInfo(item.id).type))
				return true;

		return false;
	}

	bool Base::hasOnlyLossless(const Bucket& bucket) const
	{
		for(const Bucket::Item& item : bucket.items())
			if (!isLossType(imageInfo(item.id).type))
				return false;

		return true;
	}

	FileSize Base::fileSize(const Bucket& bucket) const
	{
		if (bucket.items().empty())
			return 0;

		if (bucket.isUpToDate(CompressionCost))
		{
			return bucket.m_fileSize;
		}
		else
		{
			bucket.setUpToDate(CompressionCost, true);
			bucket.m_fileSize = compress(bucket);
			bucket.m_compressionCost = costFromFileSize(bucket.m_fileSize);
			return bucket.m_fileSize;
		}
	}

	FileSize Base::fileSize(const Buckets& buckets) const
	{
		FileSize result = 0;
		for (const Bucket& bucket : buckets)
		{
			result += fileSize(bucket);
		}
		return result;
	}

	Cost Base::cost(const Buckets& buckets, CostType costType) const
	{
		if (buckets.empty())
			return 0;

		if (costType == GeometryCost)
		{
			Cost result = 0;
			for (const Bucket& bucket : buckets)
			{
				result += cost(bucket, costType);
			}
			return result;
		}

		if (costType == CompressionCost)
		{
			vector<Cost> costs;
			costs.reserve(buckets.size());
			for (Buckets::const_iterator itr = buckets.begin(); itr != buckets.end(); ++itr)
			{
				costs.push_back(cost(*itr, CompressionCost));
			}
			return costFunction(costs);
		}

		return inifnity();
	}

	Cost Base::costFunction(const vector<Cost> costs) const
	{
		if (costs.size() < 1)
			return inifnity();

		int lastC = (int)speedMbpsTable.size() - 1;

		int startC = costs.size();
		if (startC > lastC) { startC = lastC; }
		double v1 = speedMbps(1);
		Cost startupCost = 1000 * startupMs;
		Cost finalCost = 100000000;

		for (int c = startC; c > 0; c--) {
			Cost sumCost = 0;
			Cost maxCost = 0;
			double vc = speedMbps(c);
			double vcc = vc / c;
			for (vector<Cost>::const_iterator itr = costs.begin(); itr != costs.end(); ++itr)
			{
				Cost singleCost = *itr;
				Cost singleCostWithoutStartup = singleCost - startupCost; // singleCostWithoutStartup = FileSizeInMegabits / v1
				double fileSizeInMegabits = singleCostWithoutStartup * v1;
				sumCost += (double)startupCost + fileSizeInMegabits / vcc;
				maxCost = max<Cost>(maxCost, (double)startupCost + fileSizeInMegabits / vcc);
			}
			sumCost = sumCost / c;
			finalCost = min(finalCost, max(sumCost, maxCost));
		}
		return finalCost;
	}

	Cost Base::cost(const Bucket& bucket, CostType costType) const
	{
		switch (costType)
		{
		case GeometryCost:
			return geometryCost(bucket);

		case CompressionCost:
			return compressionCost(bucket);

		default:
			return Cost::infinity();
		}
	}

	string Base::geometryMoveName(const Bucket& bucket) const
	{
		return bucket.m_geometryMoveName;
	}

	string Base::compressionMoveName(const Bucket& bucket) const
	{
		return bucket.m_compressionMoveName;
	}

	void Base::setGeometryMoveName(const Bucket& bucket, const string& name) const
	{
		bucket.m_geometryMoveName = name;
	}

	void Base::setCompressionMoveName(const Bucket& bucket, const string& name) const
	{
		bucket.m_compressionMoveName = name;
	}

	Cost Base::geometryCost(const Bucket::Items& items) const
	{
		return (Cost)geometryWaste(items) + (Cost)(area(items) / 10);
	}

	Cost Base::geometryCost(const Bucket& bucket) const
	{
		if (bucket.items().empty())
			return Cost::infinity();

		if (bucket.isUpToDate(GeometryCost))
		{
			return bucket.m_geometryCost;
		}
		else
		{
			bucket.m_geometryCost = geometryCost(bucket.items());
			bucket.setUpToDate(GeometryCost, true);
			return bucket.m_geometryCost;
		}
	}

	Area Base::area(const Geometry& _geometry) const //non cached
	{
		return ((Area)_geometry.width) * ((Area)_geometry.height);
	}

	Area Base::area(const Bucket::Items& items) const //non cached
	{
		Geometry _geometry = geometry(items);
		return ((Area)_geometry.width) * ((Area)_geometry.height);
	}

	Area Base::area(const Buckets& buckets) const //non cached
	{
		Area result = 0;
		for (const Bucket& bucket : buckets)
		{
			result += area(bucket);
		}
		return result;
	}

	Area Base::area(const Bucket& bucket) const //non cached
	{
		Geometry _geometry = geometry(bucket);

		return ((Area)_geometry.width) * ((Area)_geometry.height);
	}

	Cost Base::compressionCost(const Bucket& bucket) const
	{
		if (bucket.items().empty())
			return Cost::infinity();

		if (bucket.isUpToDate(CompressionCost))
		{
			return bucket.m_compressionCost;
		}
		else
		{
			bucket.setUpToDate(CompressionCost, true);
			bucket.m_fileSize = compress(bucket);
			bucket.m_compressionCost = costFromFileSize(bucket.m_fileSize);
			return bucket.m_compressionCost;
		}
	}

	FileSize Base::compress(const Bucket& bucket, const string& outputFilePath) const
	{
		FileSize result = -1;

		const ImageInfo * singleLossy = 0;

		if (bucket.items().size() == 1)
		{
			const Bucket::Item& item = bucket.items().front();
			if (isValid(item))
			{
				const ImageInfo& info = imageInfo(item);
				if (isLossType(info.type))
				{
					singleLossy = &info;
				}
			}
		}

		if (bucket.type() == None && singleLossy)
		{
			result = singleLossy->afterFileSize;

			if (!outputFilePath.empty())
			{
				const string fullFileName = pathJoin(inputDir, singleLossy->fileName);
				fstream from(fullFileName, fstream::in | fstream::binary);
				fstream to(outputFilePath.c_str(), fstream::out | fstream::trunc | fstream::binary);
				to << from.rdbuf();
			}
		}
		else if (bucket.type() != None && singleLossy && isLossType(bucket.type()))
		{
			result = -1;
		}
		else if (bucket.type() != None)
		{
			Geometry _geometry = geometry(bucket);
			Image img = Image(Magick::Geometry(_geometry.width, _geometry.height), background(bucket));
			if (hasTransparency(bucket)) {
				img.transparent(Color(0, 0, 0, 0));
			}
			for (const Bucket::Item& item : bucket.items())
			{
				if (isValid(item))
				{
					const Image& img2 = image(item);
					img.composite(img2, item.x, item.y, MagickCore::CopyCompositeOp);
				}
			}
			
			img.type(Magick::OptimizeType);
			img.strip();
			if (bucket.interlaceType())
				img.interlaceType((MagickCore::InterlaceType)bucket.interlaceType());

			if (typeHasQuality(bucket.type()))
				img.quality(bucket.quality());
			

			img.magick(bucket.type().toString());

			if (!outputFilePath.empty())
			{
				img.write(outputFilePath);
				result = img.fileSize();
			}
			else
			{
				Magick::Blob blob;
				img.write(&blob);
				result = blob.length();
			}
		}

		return result;
	}

	Cost Base::costFromFileSize(FileSize fileSize) const
	{
		if (fileSize <= 0)
		{
			return Cost::infinity();
		}
		else
		{
			return (Cost::Type)(1000 * startupMs) + (Cost::Type)((1000.0 * 1000.0) * 1 / (1024.0 * 1024.0) * (double)(8 * fileSize) / speedMbps(1));
		}
	}

	void Base::saveBuckets(const Buckets& buckets) const
	{
		ensureFileNames(buckets);
		for (const Bucket& bucket : buckets)
			compress(bucket, pathJoin(outputDir, bucket.m_fileName));
	}

	struct SpritePackException : public exception
	{
		SpritePackException(const string& what = "") throw()
			: m_what(what)
		{}

		virtual ~SpritePackException() throw()
		{}
		
		virtual const char * what() const throw()
		{
			return m_what.c_str();
		}

		string m_what;
	};

	void Base::throwError(const string& s) const
	{
		Log() << "EXCEPTION:" << s;
		throw SpritePackException(s);
	}

	bool Base::isCollision(const Bucket::Items& items) const
	{
		for (Bucket::Items::const_iterator itr1 = items.begin(); itr1 != items.end(); ++itr1)
		{
			for (Bucket::Items::const_iterator itr2 = ++Bucket::Items::const_iterator(itr1); itr2 != items.end(); ++itr2)
			{
				int left1 = itr1->x;
				int top1 = itr1->y;
				int right1 = itr1->x + imageInfo(itr1->id).width;
				int bottom1 = itr1->y + imageInfo(itr1->id).height;

				int left2 = itr2->x;
				int top2 = itr2->y;
				int right2 = itr2->x + imageInfo(itr2->id).width;
				int bottom2 = itr2->y + imageInfo(itr2->id).height;

				if (left1 < right2 && left2 < right1 && top1 < bottom2 && top2 < bottom1)
					return true;
			}
		}
		return false;
	}

	bool Base::isCollision(const Bucket& bucket) const
	{
		return isCollision(bucket.items());
	}

	Move Base::createNullMove(Cost cost)
	{
		return Move(cost);
	}

	string Base::bucketExt(const Bucket& bucket) const
	{
		string result = bucket.type().toExt();
		if (bucket.type() == None && bucket.items().size() == 1)
		{
			const ImageInfo& info = imageInfo(bucket.items().front());
			result = info.type.toExt();
		}
		return result;
	}

	void Base::generateFileNames(const Buckets& buckets) const
	{
		int id = 1;
		for (const Bucket& bucket : buckets)
		{
			bucket.m_fileName = toString(id) + bucketExt(bucket);
			++id;
		}
	}

	void Base::ensureFileNames(const Buckets& buckets) const
	{
		bool hasEmptyFileName = false;
		for (const Bucket& bucket : buckets)
		{
			if (bucket.fileName().empty())
			{
				hasEmptyFileName = true;
				break;
			}
		}

		if (hasEmptyFileName)
			generateFileNames(buckets);
	}

	void Base::showInputParamsInfo() const
	{
		Log() << "inputDir:" << inputDir;
		Log() << "outputDir:" << outputDir;
		Log() << "jpegQuality:" << jpegQuality;
		Log() << "pngCompressionLevel:" << pngCompressionLevel;
		Log() << "pngFilter:" << pngFilter;
		Log() << "k:" << k;
		Log() << "startupMs:" << startupMs;
		Log() << "allowLossyTypeChange:" << allowLossyTypeChange;
		Log() << "useBL:" << useBL;
		Log() << "useMBL:" << useMBL;
		Log() << "useVHTL:" << useVHTL;
		Log() << "useTwoFit:" << useTwoFit;
		Log() << "useFFDH:" << useFFDH;
		Log() << "useNFDH:" << useNFDH;
		Log() << "useBFDH:" << useBFDH;
		Log() << "postProcessCmd:" << postprocessCmd;
		for (int i = 1; i < speedMbpsTable.size(); ++i)
		{
			Log() << "SpeedMbps[" + toString(i) + "]:" << speedMbpsTable[i];
		}
	}

	void Base::showStats(const Buckets& buckets) const
	{
		
		const string costGain = toString(100.0 * (inputCost - cost(buckets)) / inputCost) + "%";
		const string costInput = inputCost.toString(CompressionCost);
		const string costOutput = cost(buckets).toString(CompressionCost);

		const string countGain = toString(100.0 * (imageInfos.size() - buckets.size()) / imageInfos.size()) + "%";
		const string countInput = toString(imageInfos.size());
		const string countOutput = toString(buckets.size());

		const string fileSizeGain = toString(100.0 * (inputFileSize - fileSize(buckets)) / inputFileSize) + "%";
		const string fileSizeInput = toString(inputFileSize);
		const string fileSizeOutput = toString(fileSize(buckets));

		const string areaGain = toString(100.0 * (inputArea - area(buckets)) / inputArea) + "%";
		const string areaInput = toString(inputArea);
		const string areaOutput = toString(area(buckets));
		
		/*
		Log() << "-- BUCKETS STATS --";

		for (const Bucket& bucket : buckets)
		{
			Log() << "Filename:" << bucket.fileName();
			Log() << "Compression:" << compressionMoveName(bucket);
			Log() << "Geometry:" << geometryMoveName(bucket);
			Log() << "Area:" << area(bucket);
			Log() << "Cost:" << cost(bucket, CompressionCost).toString(CompressionCost);
			Log() << "Geometry waste:" << cost(bucket, GeometryCost).toString(GeometryCost) << "(" + toString(100.0 * cost(bucket, GeometryCost) / area(bucket)) + "%)";
			Log() << "FileSize:" << fileSize(bucket);
			Log();
		}

		Log() << "-- MOVES STATS --";
		for (map<string, int>::const_iterator itr = geometryMovesStats.begin(); itr != geometryMovesStats.end(); ++itr)		
			Log() << itr->first + ":" << itr->second;		

		for (map<string, int>::const_iterator itr = compressionMovesStats.begin(); itr != compressionMovesStats.end(); ++itr)
			Log() << itr->first + ":" << itr->second;
		Log();

		Log() << "-- CLASSES STATS --";
		for (map<string, int>::const_iterator itr = classesStats.begin(); itr != classesStats.end(); ++itr)
			Log() << itr->first + ":" << itr->second;
		Log();
		*/

		Log() << "-- GLOBAL STATS --";
		Log() << "Used k: " + toString(usedK) + " (" + toString((100.0 * usedK)/imageInfos.size()) + "%)";
		Log() << "Classes count: " + toString(classesCount);
		Log() << "Cost gain: " + costGain + " (" + costInput + " -> " + costOutput + ")";
		Log() << "File Count gain: " + countGain + " (" + countInput + " -> " + countOutput + ")";
		Log() << "File Size gain: " + fileSizeGain + " (" + fileSizeInput + " -> " + fileSizeOutput + ")";
		Log() << "Area gain: " + areaGain + " (" + areaInput + " -> " + areaOutput + ")";
	}

	string Base::cssLine(const Bucket::Item& item, const string& filePath) const
	{
		const ImageInfo& info = imageInfo(item);
		const string className = info.cssClassName;
		const string x = item.x == 0 ? "0" : "-" + toString(item.x) + "px";
		const string y = item.y == 0 ? "0" : "-" + toString(item.y) + "px";
		const string width = toString(info.width) +"px";
		const string height = toString(info.height) + "px";

		return "." + className + " { display: inline-block; background: url('" + filePath + "') " + x + " " + y + " no-repeat; width: " + width + "; height: " + height + " }";
	}

	Base::IncreasingClassNameCmp::IncreasingClassNameCmp(const Base * _base)
		: base(_base)
	{}

	bool Base::IncreasingClassNameCmp::operator()(Id a, Id b) const
	{
		if (!base)
			return false;

		return base->imageInfo(a).cssClassName < base->imageInfo(b).cssClassName;
	}

	string Base::prettyFileSize(const FileSize& fileSize) const
	{
		FileSize kb = fileSize;
		string s = toString(kb);
		string result;
		for (int i = 0; i < (int)s.size(); ++i)
		{
			result += s[i];
			if ((s.size() - i - 1) % 3 == 0)
				result += " ";
		}
		return trim(result) + "B";
	}

	void Base::generateHtml(const Buckets& buckets, const BeforePostProcessInfo& beforePostProcess) const
	{
		const string time = Time::toString();

		const string cssFile = "style.css";

		const string costGain = toString(100.0 * (inputCost - cost(buckets)) / inputCost) + "%";
		const string costInput = inputCost.toString(CompressionCost);
		const string costOutput = cost(buckets).toString(CompressionCost);

		const string countGain = toString(100.0 * (imageInfos.size() - buckets.size()) / imageInfos.size()) + "%";
		const string countInput = toString(imageInfos.size());
		const string countOutput = toString(buckets.size());

		const string fileSizeGain = toString(100.0 * (inputFileSize - fileSize(buckets)) / inputFileSize) + "%";
		const string fileSizeInput = prettyFileSize(inputFileSize);
		const string fileSizeOutput = prettyFileSize(fileSize(buckets));

		const string areaGain = toString(100.0 * (inputArea - area(buckets)) / inputArea) + "%";
		const string areaInput = toString(inputArea);
		const string areaOutput = toString(area(buckets));

		fstream file(pathJoin(outputDir, "index.html"), fstream::trunc | fstream::binary | fstream::out);

		file << "<!doctype html>" << endl;
		file << "<head>" << endl;
		file << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />" << endl;
		file << "<title>SpritePack</title>" << endl;
		file << "<style type=\"text/css\">" << endl;
		file << ".bucket { border-top: 1px solid black }" << endl;
		file << ".bucket:nth-child(1) { border-top: none }" << endl;
		file << "table, .buckets { margin: 18px 5px; border: 1px solid black; padding: 14px 2px 2px 2px;  }" << endl;
		file << ".tt, .tt2 { border: 1px solid black; position: absolute; background: white; z-index: 1; -webkit-border-radius: 5px; -moz-border-radius: 5px; border-radius: 5px; padding: 3px 15px; margin-top: -13px; }" << endl;
		file << ".tt { left: 20px; }" << endl;
		file << ".tt2 { right: 20px; }" << endl;
		file << "table td + td { text-align: right; }" << endl;
		file << "</style>" << endl;
		file << "<link rel=\"stylesheet\" type=\"text/css\" href=\"" + cssFile + "\"></link>" << endl;
		file << "</head>" << endl;
		file << "<body>" << endl;


		file << "<div style=\"float: right\">" << endl;
		file << "<div class=\"tt2\">Parameters</div>" << endl;
		file << "<table style=\"margin: 0px 5px\">" << endl;
		file << "<thead>" << endl;
		file << "<th>Param</th>" << endl;
		file << "<th>Value</th>" << endl;
		file << "</thead>" << endl;
		file << "<tbody>" << endl;
		file << "<tr><td>Input</td><td>" << inputDir << "</td></tr>" << endl;
		file << "<tr><td>Output</td><td>" << outputDir << "</td></tr>" << endl;
		file << "<tr><td>Jpeg quality</td><td>" << jpegQuality << "</td></tr>" << endl;
		file << "<tr><td>Png compression</td><td>" << pngCompressionLevel << "</td></tr>" << endl;
		file << "<tr><td>Png filter</td><td>" << (pngFilter == -1 ? "Not specified" : toString(pngFilter)) << "</td></tr>" << endl;
		file << "<tr><td>K</td><td>" << k << "</td></tr>" << endl;
		file << "<tr><td>usedK</td><td>" << toString(usedK) + " (" + toString((100.0 * usedK) / imageInfos.size()) + "%)" << "</td></tr>" << endl;
		file << "<tr><td>classesCount</td><td>" << classesCount << "</td></tr>" << endl;
		file << "<tr><td>Startup</td><td>" << startupMs << "ms</td></tr>" << endl;
		// file << "<tr><td>SpeedForOne</td><td>" << speedMbps1 << "Mbps</td></tr>" << endl;
		file << "<tr><td>Allow lossy type change</td><td>" << toString(allowLossyTypeChange) << "</td></tr>" << endl;
		file << "<tr><td>useBL</td><td>" << toString(useBL) << "</td></tr>" << endl;
		file << "<tr><td>useMBL</td><td>" << toString(useMBL) << "</td></tr>" << endl;
		file << "<tr><td>useVHTL</td><td>" << toString(useVHTL) << "</td></tr>" << endl;
		file << "<tr><td>useTwoFit</td><td>" << toString(useTwoFit) << "</td></tr>" << endl;
		file << "<tr><td>useFFDH</td><td>" << toString(useFFDH) << "</td></tr>" << endl;
		file << "<tr><td>useNFDH</td><td>" << toString(useNFDH) << "</td></tr>" << endl;
		file << "<tr><td>useBFDH</td><td>" << toString(useBFDH) << "</td></tr>" << endl;
		file << "<tr><td>postProcessCmd</td><td>" << postprocessCmd << "</td></tr>" << endl;
		for (int i = 1; i < speedMbpsTable.size(); ++i)
		{
			file << "<tr><td>SpeedMbps[" + toString(i) + "]</td><td>" << speedMbpsTable[i] << "Mbps</td></tr>" << endl;
		}
		file << "</tbody>" << endl;
		file << "</table>" << endl;
		file << "</div>" << endl;



		file << "<div class=\"tt\">Results summary</div>" << endl;
		file << "<table>" << endl;
		file << "<thead>" << endl;

		if (!postprocessCmd.empty())
		{
			file << "<th>&nbsp;</th>" << endl;
			file << "<th>Before</th>" << endl;
			file << "<th>Spritepack</th>" << endl;
			file << "<th>Gain</th>" << endl;
			file << "<th>Postprocess</th>" << endl;
			file << "<th>Gain</th>" << endl;
			file << "<th>Total gain</th>" << endl;
		}
		else
		{
			file << "<th>&nbsp;</th>" << endl;
			file << "<th>Before</th>" << endl;
			file << "<th>Spritepack</th>" << endl;
			file << "<th>Gain</th>" << endl;
		}


		file << "</thead>" << endl;
		file << "<tfoot>" << endl;
		file << "<td>Running time</td>" << endl;
		if (!postprocessCmd.empty())
		{
			file << "<td colspan=\"6\">" + time + "</td>" << endl;
		}
		else
		{
			file << "<td colspan=\"3\">" + time + "</td>" << endl;
		}
		file << "</tfoot>" << endl;
		file << "<tbody>" << endl;
		if (!postprocessCmd.empty())
		{
			const string postProcessCostGain = toString(100.0 * (beforePostProcess.first - cost(buckets, CompressionCost)) / cost(buckets, CompressionCost)) + "%";
			const string postProcessFileSizeGain = toString(100.0 * (beforePostProcess.second - fileSize(buckets)) / fileSize(buckets)) + "%";
			const string beforePostProcessCost = beforePostProcess.first.toString(CompressionCost);
			const string beforePostProcessCostGain = toString(100.0 * (inputCost - beforePostProcess.first) / inputCost) + "%";
			const string beforePostProcessFileSize = prettyFileSize(beforePostProcess.second);
			const string beforePostProcessFileSizeGain = toString(100.0 * (inputFileSize - beforePostProcess.second) / inputFileSize) + "%";

			file << "<tr><td>Cost</td><td>" + costInput + "</td><td>" + beforePostProcessCost + "</td><td>" + beforePostProcessCostGain + "</td><td>" + costOutput + "</td><td>" + postProcessCostGain + "</td><td>" + costGain + "</td></tr>" << endl;
			file << "<tr><td>File count</td><td>" + countInput + "</td><td>" + countOutput + "</td><td>" + countGain + "</td></tr>" << endl;
			file << "<tr><td>File size</td><td>" + fileSizeInput + "</td><td>" + beforePostProcessFileSize + "</td><td>" + beforePostProcessFileSizeGain + "</td><td>" + fileSizeOutput + "</td><td>" + postProcessFileSizeGain + "</td><td>" + fileSizeGain + "</td></tr>" << endl;
			file << "<tr><td>Area</td><td>" + areaInput + "px" + "</td><td>" + areaOutput + "px" + "</td><td>" + areaGain + "</td></tr>" << endl;
		}
		else
		{
			file << "<tr><td>Cost</td><td>" + costInput + "</td><td>" + costOutput + "</td><td>" + costGain + "</td></tr>" << endl;
			file << "<tr><td>File count</td><td>" + countInput + "</td><td>" + countOutput + "</td><td>" + countGain + "</td></tr>" << endl;
			file << "<tr><td>File size</td><td>" + fileSizeInput + "</td><td>" + fileSizeOutput + "</td><td>" + fileSizeGain + "</td></tr>" << endl;
			file << "<tr><td>Area</td><td>" + areaInput + "px" + "</td><td>" + areaOutput + "px" + "</td><td>" + areaGain  + "</td></tr>" << endl;
		}
		file << "</tbody>" << endl;
		file << "</table>" << endl;

		file << "<div class=\"tt\">Input images</div>" << endl;
		file << "<table>" << endl;
		file << "<thead>" << endl;
		file << "<th>Filename</th>" << endl;
		file << "<th>Type</th>" << endl;
		file << "<th>Palette</th>" << endl;
		file << "<th>Has transp.</th>" << endl;
		file << "<th>Css class</th>" << endl;
		file << "<th>Before file size</th>" << endl;
		file << "<th>After file size</th>" << endl;
		file << "</thead>" << endl;
		file << "<tbody>" << endl;
		for (const ImageInfo& info : imageInfos)
		{
			file << "<tr><td>" << info.fileName << "</td><td>" << info.type.toString() << "</td><td>" << info.palette.toString() << "</td><td>" << toString(info.palette.hasTransparency()) << "</td><td>" << info.cssClassName << "</td><td>" << prettyFileSize(info.beforeFileSize) << "</td><td>" << prettyFileSize(info.afterFileSize) << "</td></tr>" << endl;
		}
		file << "</tbody>" << endl;
		file << "</table>" << endl;

		file << "<div class=\"tt\">Image class assignment</div>" << endl;
		file << "<table>" << endl;
		file << "<thead>" << endl;
		file << "<th>Class name</th>" << endl;
		file << "<th>Class count</th>" << endl;
		file << "</thead>" << endl;
		file << "<tbody>" << endl;
		for (map<string, int>::const_iterator itr = classesStats.begin(); itr != classesStats.end(); ++itr)
			file << "<tr><td>" << itr->first << "</td><td>" << itr->second << "</td></tr>" << endl;
		file << "</tbody>" << endl;
		file << "</table>" << endl;


		file << "<div class=\"tt\">Used methods</div>" << endl;
		file << "<table>" << endl;
		file << "<thead>" << endl;
		file << "<th>Type</th>" << endl;
		file << "<th>Usage</th>" << endl;
		file << "</thead>" << endl;
		file << "<tbody>" << endl;
		for (map<string, int>::const_iterator itr = geometryMovesStats.begin(); itr != geometryMovesStats.end(); ++itr)		
			file << "<tr><td>" << itr->first << "</td><td>" <<  itr->second << "</td></tr>" << endl;

		for (map<string, int>::const_iterator itr = compressionMovesStats.begin(); itr != compressionMovesStats.end(); ++itr)
			file << "<tr><td>" << itr->first << "</td><td>" <<  itr->second << "</td></tr>" << endl;
		file << "</tbody>" << endl;
		file << "</table>" << endl;



		file << "<div class=\"tt\">Generated sprites</div>" << endl;
		file << "<table>" << endl;
		file << "<thead>" << endl;
		file << "<th>Filename</th>" << endl;
		file << "<th>Compression</th>" << endl;
		file << "<th>Geometry</th>" << endl;
		file << "<th>Area</th>" << endl;
		file << "<th>Cost</th>" << endl;
		file << "<th>Geometry waste</th>" << endl;
		file << "<th>File size</th>" << endl;
		file << "</thead>" << endl;
		file << "<tbody>" << endl;
		for (const Bucket& bucket : buckets)
		{
			file << "<tr>" << endl;
			file << "<td>" + bucket.fileName() + "</td>" << endl;
			file << "<td>" + compressionMoveName(bucket) + "</td>" << endl;
			file << "<td>" + geometryMoveName(bucket) + "</td>" << endl;
			file << "<td>" + toString(area(bucket)) + "px" + "</td>" << endl;
			file << "<td>" + cost(bucket, CompressionCost).toString(CompressionCost) + "</td>" << endl;
			file << "<td>" + toString(geometryWaste(bucket)) + "px" + " (" + toString(100.0 * geometryWaste(bucket) / area(bucket)) + "%)</td>" << endl;
			file << "<td>" + prettyFileSize(fileSize(bucket)) + "</td>" << endl;
			file << "</tr>" << endl;

		}
		file << "</tbody>" << endl;
		file << "</table>" << endl;

		file << "<div class=\"tt\">CSS rules verification</div>" << endl;
		file << "<div class=\"buckets\">" << endl;
		for (const Bucket& bucket : buckets)
		{
			file << "<div class=\"bucket\">" + bucket.fileName() + "<br>" << endl;
			Bucket::Items items = bucket.items();
			sort(items.begin(), items.end(), IncreasingClassNameCmp(this));
			for (const Bucket::Item& item : items)
			{
				const ImageInfo& info = imageInfo(item);
				file << "<span class=\"" + info.cssClassName + "\" alt=\"" + info.fileName + "\"></span>" << endl;
			}
			file << "</div>" << endl;
		}
		file << "</div>" << endl;


		file << "</body>" << endl;
		file << "</html>" << endl;
	}

	Number Base::geometryWaste(const Bucket::Items& items) const
	{
		Number r = 0;
		for (const Bucket::Item& item : items)
		{
			if (isValid(item))
			{
				const ImageInfo& info = imageInfo(item);
				r += area(info.geometry());
			}
		}
		return area(items) - r;
	}

	Number Base::geometryWaste(const Bucket& bucket) const
	{
		return geometryWaste(bucket.m_items);
	}

	Number Base::geometryWaste(const Buckets& buckets) const
	{
		Number r = 0;
		for (const Bucket& bucket : buckets)
		{
			r += geometryWaste(bucket);
		}
		return 0;
	}

	void Base::generateCss(const Buckets& buckets) const
	{
		ensureFileNames(buckets);

		fstream file(pathJoin(outputDir, "style.css"), fstream::trunc | fstream::binary | fstream::out);

		int i = 0;
		for (const Bucket& bucket : buckets)
		{
			Bucket::Items items = bucket.items();
			sort(items.begin(), items.end(), IncreasingClassNameCmp(this));
			for (const Bucket::Item& item : items)
			{
				file << cssLine(item, bucket.fileName()) << endl;
			}
			++i;
		}
	}

	void Base::generateHtmlCss(const Buckets& buckets, const BeforePostProcessInfo& beforePostProcess) const
	{
		generateCss(buckets);
		generateHtml(buckets, beforePostProcess);
	}

	FileSize Base::fileSizeFromFileName(const string& fullFileName) const
	{
		std::ifstream in(fullFileName, std::ifstream::in | std::ifstream::binary);		
		in.seekg(0, std::ifstream::end);
		return in.tellg();
	}

	void Base::makePostprocess(const Buckets& buckets) const
	{
		return;
		if (!postprocessCmd.empty())
		{
			int x = system((postprocessCmd + " \"" + outputDir + "\"").c_str());

			for (const Bucket& bucket : buckets)
			{
				bucket.setUpToDate(CompressionCost, true);
				bucket.m_fileSize = fileSizeFromFileName(pathJoin(outputDir, bucket.m_fileName));
				bucket.m_compressionCost = costFromFileSize(bucket.m_fileSize);
				bucket.m_compressionMoveName = "PostProcess";
			}
		}
	}

} /* SpritePack */