/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef BASE_H
#define BASE_H

#include "utils.h"
#include "cost.h"
#include "bucket.h"
#include "type.h"
#include "move.h"
#include "direction.h"
#include "imageinfo.h"
#include "costtype.h"
#include "geometry.h"

namespace SpritePack
{
	using Magick::Image;

	using Magick::Color;

	typedef vector<ImageInfo> ImageInfos;

	typedef vector<Image> Images;

	typedef pair<Cost, FileSize> BeforePostProcessInfo;

	class Base
	{
	public:
		string inputDir;

		string outputDir;

		string postprocessCmd;

		Quality jpegQuality;  // 1 - 100

		Quality pngCompressionLevel; // 1 - 9

		int pngFilter; // 1 - 9

		int startupMs;

		bool allowLossyTypeChange;

		string k;

		int usedK;

		int classesCount;

		Cost inputCost;

		FileSize inputFileSize;

		Area inputArea;

		double loadFinishedTime;

		bool useBL;

		bool useMBL;

		bool useVHTL;

		bool useTwoFit;

		bool useFFDH;

		bool useNFDH;

		bool useBFDH;

		map<string, int> geometryMovesStats;
		
		map<string, int> compressionMovesStats;

		map<string, int> classesStats;

		vector<double> speedMbpsTable;

		void readK(int * numberValue, double * percentValue) const;

	public:
		Images images;

		ImageInfos imageInfos;

	public:
		const Image nullImage;

		const ImageInfo nullImageInfo;

		const Buckets nullBuckets;
	
	public:
		Base();

		virtual ~Base();

	public:
		bool isValid(Id) const;

		const ImageInfo& imageInfo(Id) const;

		const Image& image(Id) const;

		Base& operator=(const Base&);

		double speedMbps(int) const;

		static Cost inifnity();

	public:
		void openLogFile(const string& fileName = "");

		void closeLogFile();

		vector<int> pngFilters() const;

		bool isGoodExt(const string& fileName) const;

		void calcBestTypePalette(const string& fileName, Image*, Type*, Palette*, FileSize*, Quality*) const;

		void checkInputParams() const;

		void checkInputDir() const;

		void checkOutputDir() const;

		pair<ImageInfo, Image> loadImage(const string& fullFileName) const;

		void load();

		void setClassStats(const string& name, int n);

		void compressionMoveUsed(const string& name);

		void geometryMoveUsed(const string& name);

		string generateCssClassName(const string& fileName, int i = 0) const;

		bool typeHasQuality(const Type&) const;

		Geometry geometry(Id) const;

		Geometry geometry(const Bucket::Items&) const;

		Geometry geometry(const Bucket&) const;

		Color background(const Bucket&) const;

		bool hasTransparency(const Bucket&) const;

		bool isLossType(Type) const;

		bool hasType(const Bucket&, Type) const;

		bool hasOnlyLossless(const Bucket&) const;

		bool hasLossless(const Bucket&) const;

		vector<Palette> palettes() const;

		Cost cost(const Buckets&, CostType costType = CompressionCost) const;

		Cost cost(const Bucket&, CostType costType = CompressionCost) const;

		Cost costFunction(const vector<Cost> costs) const; // cost is startup + transmission time in micro seconds

		FileSize fileSize(const Bucket&) const;

		FileSize fileSize(const Buckets&) const;

		Area area(const Geometry&) const; //non cached

		Area area(const Bucket::Items&) const; //non cached

		Area area(const Buckets&) const; //non cached

		Area area(const Bucket&) const; //non cached

		Number geometryWaste(const Bucket::Items&) const;

		Number geometryWaste(const Bucket&) const;

		Number geometryWaste(const Buckets&) const;

		string geometryMoveName(const Bucket&) const;

		string compressionMoveName(const Bucket&) const;

		void setGeometryMoveName(const Bucket&, const string& name) const;

		void setCompressionMoveName(const Bucket&, const string& name) const;

		Cost geometryCost(const Bucket::Items&) const;

		Cost geometryCost(const Bucket&) const;

		Cost compressionCost(const Bucket&) const;
	
		FileSize compress(const Bucket&, const string& outputFilePath = "") const;

		Cost costFromFileSize(FileSize) const;

		FileSize fileSizeFromFileName(const string& fullFileName) const;

		void saveBuckets(const Buckets&) const;

		void showInputParamsInfo() const;

		void showStats(const Buckets&) const;

		string prettyFileSize(const FileSize& fileSize) const;

		void generateHtml(const Buckets&, const BeforePostProcessInfo& = BeforePostProcessInfo()) const;

		void generateCss(const Buckets&) const;

		void makePostprocess(const Buckets&) const;

		void generateHtmlCss(const Buckets&, const BeforePostProcessInfo& = BeforePostProcessInfo()) const;

		void generateFileNames(const Buckets&) const;

		void ensureFileNames(const Buckets&) const;

		string cssClass(Id id) const;

		string cssLine(const Bucket::Item&, const string& filePath) const;

		string htmlHeader() const;

		string bucketExt(const Bucket&) const;

		bool isCollision(const Bucket::Items&) const;

		bool isCollision(const Bucket&) const;

		void throwError(const string&) const;

		struct IncreasingClassNameCmp
		{
			const Base * base;

			IncreasingClassNameCmp(const Base *);

			bool operator()(Id, Id) const;
		};

	public:

		Move createNullMove(Cost cost);

		template<class Impl> Move createMove(Impl * = new Impl());

		template<class Impl, class A1> Move createMove(A1, Impl * = new Impl());

		template<class Impl, class A1, class A2> Move createMove(A1, A2, Impl * = new Impl());

		template<class Impl, class A1, class A2, class A3> Move createMove(A1, A2, A3, Impl * = new Impl());

		template<class Impl, class A1, class A2, class A3, class A4> Move createMove(A1, A2, A3, A4, Impl * = new Impl());

	public:
		virtual Moves compressionMoves(Bucket*) = 0;

		virtual Moves directionGeometryMoves(Bucket*, Direction) = 0;

		virtual Moves directionLessGeometryMoves(Bucket*) = 0;

		virtual void make() = 0;
	};

	/* implementation */

	template<class Impl> Move Base::createMove(Impl * impl)
	{
		impl->setBase(this);
		return Move(impl);
	}

	template<class Impl, class A1> Move Base::createMove(A1 a1, Impl * impl)
	{
		impl->setBase(this);
		impl->setInput(a1);		
		return Move(impl);
	}

	template<class Impl, class A1, class A2> Move Base::createMove(A1 a1, A2 a2, Impl * impl)
	{
		impl->setBase(this);
		impl->setInput(a1, a2);
		return Move(impl);
	}

	template<class Impl, class A1, class A2, class A3> Move Base::createMove(A1 a1, A2 a2, A3 a3, Impl * impl)
	{
		impl->setBase(this);
		impl->setInput(a1, a2, a3);
		return Move(impl);
	}

	template<class Impl, class A1, class A2, class A3, class A4> Move Base::createMove(A1 a1, A2 a2, A3 a3, A4 a4, Impl * impl)
	{
		impl->setBase(this);
		impl->setInput(a1, a2, a3, a4);
		return Move(impl);
	}

} /* namespace SpritePack */

#endif /* BASE_H */