/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "bl.h"
#include "log.h"
#include "base.h"

namespace SpritePack {

	BL::S::S(const Point& _point, int _left, int _right)
		: point(_point), left(_left), right(_right)
	{}

	BL::BL()
		: morePrecisie(false), manyWidths(false)
	{}

	vector<int> BL::areaWidths() const
	{
		if (!base)
			return vector<int>();

		vector<int> widths;

		for(const ImageInfo& info : base->imageInfos)
			widths.push_back(direction == Vertical ? info.width : info.height);

		sort(widths.begin(), widths.end(), greater<int>());

		vector<int> result;

		int acc = 0;
		for (int w : widths)
		{
			acc += w;
			result.push_back(acc);
		}

		return result;
	}
	
	void BL::setInput(Bucket * bucket, Direction direction, bool _morePrecisie, bool _manyWidths)
	{
		DirectionGeometryBucketMove::setInput(bucket, direction);
		morePrecisie = _morePrecisie;
		manyWidths = _manyWidths;
	}

	Cost BL::calc()
	{
		if (!manyWidths)
		{
			return calc(maxNormalWidth());
		}
		else
		{
			Bucket bestResult = result;
			Cost bestCost = Cost::infinity();
			for (int areaW : areaWidths())
			{
				Cost cost = calc(areaW);
				if (cost < bestCost)
				{
					bestCost = cost;
					bestResult = result;
				}
				result.clear();
			}
			result = bestResult;
			return bestCost;
		}
	}

	Cost BL::calc(const int areaW)
	{
		vector<Id> ids = getImageIds();
		sort(ids.begin(), ids.end(), DecreasingNormalWidthCmp(this));

		typedef vector<S> V;
		typedef V::iterator Itr;

		V v;
		v.push_back(S(Point(0, 0), 0, areaW));

		list<Id> idsList(ids.begin(), ids.end());
		while (!idsList.empty())
		{
			Itr insertItr = v.end();
			int bestY = numeric_limits<int>::max();

			//insertItr
			list<Id>::iterator idsListBestItr = idsList.begin();
			for (list<Id>::iterator idsListItr = idsList.begin(); idsListItr != idsList.end(); ++idsListItr)
			{
				const int w = normalWidth(*idsListItr);
				bool wasBetter = false;
				
				//cache friendly
				for (Itr itr = v.begin(); itr != v.end(); ++itr)
				{
					if (itr->left + itr->right >= w && itr->point.y < bestY)
					{
						insertItr = itr;
						bestY = itr->point.y;
						wasBetter = true;
					}
				}
				if (wasBetter)
					idsListBestItr = idsListItr;

				if (!morePrecisie)
					break;
			}

			const Id item = *idsListBestItr;
			idsList.erase(idsListBestItr);

			//init
			const int itemW = normalWidth(item);
			const int itemH = normalHeight(item);

			Point insertPoint(insertItr->point.x - insertItr->left, insertItr->point.y);
			const int insertY1(insertPoint.y);
			const int insertY2(insertY1 + itemH);
			const int insertX1(insertPoint.x);
			const int insertX2(insertX1 + itemW);

			Itr firstCovered = v.begin();
			Itr firstNotCovered = v.end();

			S s1(Point(insertPoint.x, insertPoint.y + itemH), insertPoint.x, areaW - insertPoint.x);
			S s2(Point(insertPoint.x + itemW, insertPoint.y + itemH), insertPoint.x + itemW, areaW - insertPoint.x - itemW);
			S s3(Point(insertPoint.x + itemW, insertPoint.y), 0, areaW - insertPoint.x - itemW);

			//firstCovered
			firstCovered = insertItr;
			for (Itr itr = v.begin(); itr != v.end(); ++itr)
			{
				if (insertX1 <= itr->point.x && itr->point.x <= insertX2 && itr->point.y <= insertY2)
				{
					firstCovered = itr;
					break;					
				}
			}

			//firstNotCovered
			for (Itr itr = v.begin(); itr != v.end(); ++itr)
			{
				if (insertX2 <= itr->point.x)
				{
					firstNotCovered = itr;
					break;
				}
			}

			//correct existing items
			for (Itr itr = v.begin(); itr != v.end(); ++itr)
			{
			    if (itr->point.y < insertY2) 
				{
					if (itr->point.x <= insertX1)
					{
						int newRight = insertX1 - itr->point.x;
						if (newRight < itr->right)
							itr->right = newRight;
					}
					else if (insertX2 <= itr->point.x)
					{
						int newLeft = itr->point.x - insertX2;
						if (newLeft < itr->left)
							itr->left = newLeft;
					}
				}
			}

			//s1.left
			for (Itr itr = firstCovered; ;)
			{
				if (itr->point.y > s1.point.y)
				{
					s1.left = s1.point.x - itr->point.x;
					break;
				}

				if (itr == v.begin())
					break;

				--itr;
			}

			//s1.right
			for (Itr itr = firstNotCovered; itr != v.end(); ++itr)
			{
				if (itr->point.y > s1.point.y)
				{
					s1.right = itr->point.x - s1.point.x;
					break;
				}
			}

			//s2.left
			s2.left = s1.left + itemW;

			//s2.right
			s2.right = s1.right - itemW;

			//s3.right
			for (Itr itr = firstNotCovered; itr != v.end(); ++itr)
			{
				if (itr->point.y > s3.point.y)
				{
					s3.right = itr->point.x - s3.point.x;
					break;
				}
			}

			//insert item
			result.add(item, normalPoint(insertPoint));

			//reconstruction
			V v2;

			for (Itr itr = v.begin(); itr != firstCovered; ++itr)
				v2.push_back(*itr);

			v2.push_back(s1);
			v2.push_back(s2);
			v2.push_back(s3);

			for (Itr itr = firstNotCovered; itr != v.end(); ++itr)
				v2.push_back(*itr);

			v = v2;
		}

		checkForCollisions();

		return calcResultCost();
	}

	string BL::name() const
	{
		string result;

		if (manyWidths)
		{
			result = "BL(" + direction.toString() + ", " + toString(morePrecisie) + ", " + toString(manyWidths) + ")";
		}
		else if(morePrecisie)
		{
			result = "MBL(" + direction.toString() + ")";
		}
		else
		{
			result = "BL(" + direction.toString() + ")";
		}

		return result;
	}

	void BL::visualize() const
	{
		if (!base)
			return;

		Geometry _geometry = base->geometry(result);
		Image img = Image(Magick::Geometry(_geometry.width, _geometry.height), "white");
		for (const Bucket::Item& item : result.items())
		{
			if (base->isValid(item))
			{
				img.composite(base->image(item), item.x, item.y, MagickCore::CopyCompositeOp);			
			}
		}

		img.strokeWidth(1);
		img.strokeColor("Red");
		img.fillColor("Transparent");
		for (const Bucket::Item& item : result.items())
		{
			if (base->isValid(item))
			{
				int w = base->imageInfo(item).width;
				int h = base->imageInfo(item).height;
				img.compose(Magick::BlendCompositeOp);
				img.draw(Magick::DrawableRectangle(item.x, item.y, item.x + w, item.y + h));
			}
		}

		img.write("visualize.jpg");
	}

} // namespace SpritePack
