/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "bucket.h"
#include "imageinfo.h"

namespace SpritePack {

	Bucket::Item::Item(Id _id, int _x, int _y)
		: id(_id), x(_x), y(_y)
	{}

	Bucket::Item::Item(Id _id, const Point& point)
		: id(_id), x(point.x), y(point.y)
	{}

	Bucket::Item::operator Bucket::Id() const
	{
		return id;
	}

	Id Bucket::generateId()
	{
		static int x = 0;
		return ++x;
	}

	Bucket::Bucket()
		: m_type(None), m_quality(-1), m_interlaceType(0), m_id(0), m_fileSize(0), m_geometryCost(Cost::infinity()), m_compressionCost(Cost::infinity()), m_isUpToDateGeometry(false), m_isUpToDateCompression(false)
	{}

	Bucket::Id Bucket::id() const
	{
		if (!m_id)
			m_id = generateId();
		return m_id;
	}

	void Bucket::setId(const Bucket::Id& id) const
	{
		m_id = id;
	}

	const Bucket::Items& Bucket::items() const
	{
		return m_items;
	}

	Type Bucket::type() const
	{
		return m_type;
	}
		
	Quality Bucket::quality() const
	{
		return m_quality;
	}

	int Bucket::interlaceType() const
	{
		return m_interlaceType;
	}

	void Bucket::setInterlaceType(int i)
	{
		m_interlaceType = i;
	}

	void Bucket::setInfo(const Bucket& bucket)
	{
		modify();
		m_type = bucket.m_type;
		m_quality = bucket.m_quality;
		m_fileSize = bucket.m_fileSize;
		m_geometryCost = bucket.m_geometryCost;
		m_compressionCost = bucket.m_compressionCost;
		m_geometryMoveName = bucket.m_geometryMoveName;
		m_compressionMoveName = bucket.m_compressionMoveName;
		m_interlaceType = bucket.m_interlaceType;
	}

	void Bucket::setType(Type type)
	{
		modify();
		m_type = type;		
	}

	void Bucket::setQuality(Quality quality)
	{
		modify();
		m_quality = quality;		
	}

	void Bucket::setItems(const Bucket::Items& items)
	{
		modify();
		m_items = items;		
	}

	void Bucket::setItems(const Bucket& bucket)
	{
		modify();
		m_items = bucket.m_items;
	}

	void Bucket::modify()
	{
		m_id = 0;
		m_isUpToDateGeometry = false;
		m_isUpToDateCompression = false;
		m_fileName.clear();
	}

	void Bucket::clear()
	{
		modify();
		m_items.clear();
	}

	void Bucket::add(Id id, const Point& p)
	{
		modify();
		m_items.push_back(Bucket::Item(id, p.x, p.y));
	}

	void Bucket::add(Id id, int x, int y)
	{
		modify();
		m_items.push_back(Bucket::Item(id, x, y));
	}

	void Bucket::addItems(const Items& items)
	{
		modify();
		for(const Bucket::Item& item : items)
			m_items.push_back(item);
	}

	void Bucket::addItems(const Bucket& bucket)
	{
		addItems(bucket.items());
	}

	void Bucket::assignItems(const Items& items)
	{
		modify();
		m_items = items;
	}

	bool Bucket::isUpToDate(CostType costType) const
	{
		switch (costType)
		{
		case GeometryCost:
			return m_isUpToDateGeometry;

		case CompressionCost:
			return m_isUpToDateCompression;

		default:
			return false;
		}
	}

	void Bucket::setUpToDate(CostType costType, bool value) const
	{
		switch (costType)
		{
		case GeometryCost:
			m_isUpToDateGeometry = value;
			break;

		case CompressionCost:
			m_isUpToDateCompression = value;
			break;

		default:
			break;
		}
	}

	const string& Bucket::fileName() const
	{
		return m_fileName;
	}

	Buckets& operator+=(Buckets& buckets, const Bucket& bucket)
	{
		buckets.push_back(bucket);
		return buckets;
	}

} /* namespace SpritePack */
