/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "bucketmove.h"
#include "imageinfo.h"
#include "base.h"
#include "log.h"

namespace SpritePack {

	BucketMove::BucketMove(CostType _costType)
		: input(0), output(0), costType(_costType)
	{}

	BucketMove::~BucketMove()
	{}

	void BucketMove::setInput(Bucket * bucket)
	{
		input = bucket;
		output = bucket;
	}
	
	void BucketMove::beforeCalc()
	{
		clear();
	}

	Cost BucketMove::calcResultCost()
	{
		if (!base)
			return Cost::infinity();

		return base->cost(result, costType);
	}

	void BucketMove::clear()
	{
		if (!input)
			return;

		result.clear();
		result.setInfo(*input);
	}
		
	void BucketMove::apply()
	{
		if (!output)
			return;

		*output = result;

		switch (costType)
		{
		case GeometryCost:
			base->setGeometryMoveName(*output, name());
			break;

		case CompressionCost:
			base->setCompressionMoveName(*output, name());
			break;

		default:
			break;
		}
	}

	bool BucketMove::isImprovement()
	{
		if (!base || !input)
			return false;

		Cost before = base->cost(*input, costType);
		Cost after = cost();
		return after < before;
	}

	vector<Id> BucketMove::getImageIds() const
	{
		if (!input)
			return vector<Id>();

		vector<Id> result;
		for (const Bucket::Item& item: input->items())
		{
			result.push_back(item);
		}
		return result;
	}

	GeometryBucketMove::GeometryBucketMove()
		: BucketMove(GeometryCost)
	{}

	bool GeometryBucketMove::checkForCollisions(bool throwError)
	{
		if (!base || result.items().empty())
			return false;

		bool isCollision = base->isCollision(result);
		if (throwError && isCollision)
			base->throwError("Collision by: " + name());

		return isCollision;
	}

	CompressionBucketMove::CompressionBucketMove()
		: BucketMove(CompressionCost)
	{}

	DirectionGeometryBucketMove::DirectionGeometryBucketMove()
		: m_maxNormalWidth(-1), m_maxNormalHeight(-1), m_minNormalWidth(-1), m_minNormalHeight(-1)
	{}

	DirectionGeometryBucketMove::DecreasingNormalWidthCmp::DecreasingNormalWidthCmp(DirectionGeometryBucketMove * _move)
		: move(_move)
	{}

	bool DirectionGeometryBucketMove::DecreasingNormalWidthCmp::operator()(Id a, Id b) const
	{
		if (!move)
			return false;

		if (move->normalWidth(a) != move->normalWidth(b))
			return move->normalWidth(a) > move->normalWidth(b);
		else
			return move->base->imageInfo(a).fileName < move->base->imageInfo(b).fileName;
	}

	DirectionGeometryBucketMove::DecreasingNormalHeightCmp::DecreasingNormalHeightCmp(DirectionGeometryBucketMove * _move)
		: move(_move)
	{}

	bool DirectionGeometryBucketMove::DecreasingNormalHeightCmp::operator()(Id a, Id b) const
	{
		if (!move)
			return false;

		if (move->normalHeight(a) != move->normalHeight(b))
			return move->normalHeight(a) > move->normalHeight(b);
		else
			return move->base->imageInfo(a).fileName < move->base->imageInfo(b).fileName;
	}

	DirectionGeometryBucketMove::IncreasingNormalWidthCmp::IncreasingNormalWidthCmp(DirectionGeometryBucketMove * _move)
		: move(_move)
	{}

	bool DirectionGeometryBucketMove::IncreasingNormalWidthCmp::operator()(Id a, Id b) const
	{
		if (!move)
			return false;

		if (move->normalWidth(a) != move->normalWidth(b))
			return move->normalWidth(a) < move->normalWidth(b);
		else
			return move->base->imageInfo(a).fileName < move->base->imageInfo(b).fileName;
	}

	DirectionGeometryBucketMove::IncreasingNormalHeightCmp::IncreasingNormalHeightCmp(DirectionGeometryBucketMove * _move)
		: move(_move)
	{}

	bool DirectionGeometryBucketMove::IncreasingNormalHeightCmp::operator()(Id a, Id b) const
	{
		if (!move)
			return false;

		if (move->normalHeight(a) != move->normalHeight(b))
			return move->normalHeight(a) < move->normalHeight(b);
		else
			return move->base->imageInfo(a).fileName < move->base->imageInfo(b).fileName;
	}

	void DirectionGeometryBucketMove::setInput(Bucket * bucket, Direction _direction)
	{
		GeometryBucketMove::setInput(bucket);
		direction = _direction;
		m_maxNormalWidth = -1;
		m_maxNormalHeight = -1;
	}

	Geometry DirectionGeometryBucketMove::normalGeometry(Id id) const
	{
		if (!base)
			return Geometry(0, 0);

		const ImageInfo& info = base->imageInfo(id);
		return direction == Vertical ? info.geometry() : Geometry(info.height, info.width);
	}

	Geometry DirectionGeometryBucketMove::normalGoemetry(const Geometry& geometry) const
	{
		return direction == Vertical ? geometry : Geometry(geometry.height, geometry.width);
	}

	Point DirectionGeometryBucketMove::normalPoint(int x, int y) const
	{
		return direction == Vertical ? Point(x, y) : Point(y, x);
	}

	Point DirectionGeometryBucketMove::normalPoint(const Point& p) const
	{
		return direction == Vertical ? p : Point(p.y, p.x);
	}

	int DirectionGeometryBucketMove::normalWidth(Id id) const
	{
		if (!base)
			return 0;

		const ImageInfo& info = base->imageInfo(id);
		return direction == Vertical ? info.width : info.height;
	}

	int DirectionGeometryBucketMove::normalHeight(Id id) const
	{
		if (!base)
			return 0;

		const ImageInfo& info = base->imageInfo(id);
		return direction == Vertical ? info.height : info.width;
	}

	Bucket::Items DirectionGeometryBucketMove::normalItems(const Bucket::Items& items) const
	{
		if (direction == Vertical)
		{
			return items;
		}
		else
		{
			Bucket::Items result;
			for (const Bucket::Item& item : items)		
				result.push_back(Bucket::Item(item.id, item.y, item.x));
			return result;
		}
	}

	int DirectionGeometryBucketMove::maxNormalWidth() const
	{
		if (m_maxNormalWidth >= 0)
			return m_maxNormalWidth;

		if (!input)
			return 0;

		int result = numeric_limits<int>::min();
		for (const Bucket::Item& item: input->items())
		{
			result = max(result, normalWidth(item));
		}
		return m_maxNormalWidth = result;
	}

	int DirectionGeometryBucketMove::maxNormalHeight() const
	{
		if (m_maxNormalHeight >= 0)
			return m_maxNormalHeight;

		if (!input)
			return 0;

		int result = numeric_limits<int>::min();
		for (const Bucket::Item& item: input->items())
		{
			result = max(result, normalHeight(item));
		}
		return m_maxNormalHeight = result;
	}

	int DirectionGeometryBucketMove::minNormalWidth() const
	{
		if (m_minNormalWidth >= 0)
			return m_minNormalWidth;

		if (!input)
			return 0;

		int result = numeric_limits<int>::max();
		for (const Bucket::Item& item: input->items())
		{
			result = min(result, normalWidth(item));
		}
		return m_minNormalWidth = result;
	}

	int DirectionGeometryBucketMove::minNormalHeight() const
	{
		if (m_minNormalHeight >= 0)
			return m_minNormalHeight;

		if (!input)
			return 0;

		int result = numeric_limits<int>::max();
		for (const Bucket::Item& item: input->items())
		{
			result = min(result, normalHeight(item));
		}
		return m_minNormalHeight = result;
	}

} /* namespace SpritePack */
