/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CACHE_H
#define CACHE_H

#include "thirdparty.h"

namespace SpritePack {

	template<class Key, class Value> class Cache
	{
	public:
		Cache(int limit = 100*1000);

		bool isCached(const Key&);

		const Value& insert(const Key&, const Value&);

		const Value& get(const Key&);

		void setLimit(int);

		int limit() const;

	private:
		typedef int Counter;

		typedef pair<Counter, Value> MapElement;

		typedef pair<MapElement, Key> SetElement;

		typedef map<Key, MapElement> Map;

		typedef set<SetElement> Set;

	private:
		int m_limit;

		Counter m_counter;

		Value m_nullValue;

		Map m_map;

		Set m_set;

		typename Map::iterator m_itr;

		Key m_lastKey;

		bool m_isLastKey;

	private:
		void ensureItr(const Key&);

	};


	/* implementation */

	template<class Key, class Value> Cache<Key, Value>::Cache(int limit)
		: m_limit(max(1, limit)), m_counter(0), m_nullValue(), m_itr(m_map.end()), m_isLastKey(false)
	{}

	template<class Key, class Value> bool Cache<Key, Value>::isCached(const Key& key)
	{
		ensureItr(key);
		if (m_itr != m_map.end())
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	template<class Key, class Value> const Value& Cache<Key, Value>::insert(const Key& key, const Value& value)
	{
		ensureItr(key);
		if (m_itr == m_map.end())
		{
			while (m_set.size() >= m_limit)
			{
				typename Set::iterator itr = m_set.begin();
				m_map.erase(itr->second);
				m_set.erase(itr);
			}
			++m_counter;
			MapElement me(m_counter, value);
			m_map.insert(make_pair(key, me));
			m_set.insert(SetElement(me, key));
		}
		return value;
	}

	template<class Key, class Value> const Value& Cache<Key, Value>::get(const Key& key)
	{
		ensureItr(key);
		if (m_itr != m_map.end())
		{
			if (m_itr->second.first != m_counter)
			{
				m_set.erase(SetElement(m_itr->second, key));
				m_itr->second.first = m_counter;
				m_set.insert(SetElement(m_itr->second, key));
			}
			return m_itr->second.second;
		}
		else
		{
			return m_nullValue;
		}
	}

	template<class Key, class Value> void Cache<Key, Value>::setLimit(int limit)
	{
		m_limit = max(1, limit);

		if (m_set.size() >= m_limit)
		{
			m_isLastKey = true;
		}

		while (m_set.size() >= m_limit)
		{
			typename Set::iterator itr = m_set.begin();
			m_map.erase(itr->second);
			m_set.erase(itr);
		}
	}

	template<class Key, class Value> int Cache<Key, Value>::limit() const
	{
		return m_limit;
	}

	template<class Key, class Value> void  Cache<Key, Value>::ensureItr(const Key& key)
	{
		if (m_isLastKey && !(key < m_lastKey || m_lastKey < key))
		{
			m_lastKey = key;
			m_itr = m_map.find(m_lastKey);
			m_isLastKey = true;
		}
	}

} /* namespace SpritePack */


#endif /* CACHE_H */