/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "dh.h"
#include "base.h"
#include "log.h"

namespace SpritePack {

	DH::Level::Level(int _freeSpace)
		: freeSpace(_freeSpace), height(0)
	{}

	DH::DH()
		: twoFit(false)
	{}

	void DH::setInput(Bucket * bucket, Direction direction, bool _twoFit)
	{
		DirectionGeometryBucketMove::setInput(bucket, direction);
		twoFit = _twoFit;
	}

	void DH::normalCalc(Levels * levels)
	{
		if (!levels)
			return;

		levels->clear();

		LevelsIterator itr(levels->end());

		vector<Id> ids = getImageIds();
		sort(ids.begin(), ids.end(), DecreasingNormalHeightCmp(this));

		for (Id item : ids)
		{
			const int width = normalWidth(item);

			itr = chooseLevel(levels, itr, item);

			if (itr == levels->end() || width > itr->freeSpace)
			{
				levels->push_back(Level(maxNormalWidth()));
				itr = levels->end();
				--itr;
			}

			itr->items.push_back(item);
			itr->freeSpace -= width;
			itr->height = max(itr->height, normalHeight(item));
		}
	}

	void DH::twoFitCalc(Levels * levels)
	{
		if (!levels)
			return;

		levels->clear();

		LevelsIterator itr(levels->end());

		vector<Id> ids = getImageIds();

		if (ids.empty())
			return;

		sort(ids.begin(), ids.end(), DecreasingNormalHeightCmp(this));

		list<Id> idsList(ids.begin(), ids.end());

		int minW = minNormalWidth();

		while (!idsList.empty())
		{
			list<Id>::iterator idsListItr = idsList.begin();
			list<Id>::iterator idsListItr2 = idsList.end(); 

			itr = chooseLevel(levels, itr, *idsListItr);

			if (itr != levels->end() && itr->freeSpace >= normalWidth(*idsListItr) && itr->freeSpace - normalWidth(*idsListItr) < minW) //twofit
			{
				const int minWidth = normalWidth(*idsListItr);
				const int maxWidth = itr->freeSpace;
				Id bestItem1 = -1;
				Id bestItem2 = -1;
				int bestTwoFitW = minWidth;

				for (list<Id>::iterator itr1 = idsListItr; itr1 != idsList.end(); ++itr1)
				{
					for (list<Id>::iterator itr2 = ++list<Id>::iterator(itr1); itr2 != idsList.end(); ++itr2)
					{
						int w = normalWidth(*itr1) + normalWidth(*itr2);
						if (w < maxWidth && w > bestTwoFitW)
						{
							bestTwoFitW = w;
							bestItem1 = *itr1;
							bestItem2 = *itr2;
						}
					}
				}

				if (bestItem1 != -1 && bestItem2 != -1 && bestItem1 != bestItem2)
				{
					idsListItr = find(idsList.begin(), idsList.end(), bestItem1);
					idsListItr2 = find(idsList.begin(), idsList.end(), bestItem2);
				}
			}
			else
			{
				if (itr == levels->end() || normalWidth(*idsListItr) > itr->freeSpace)
				{
					levels->push_back(Level(maxNormalWidth()));
					itr = levels->end();
					--itr;
				}
			}

			// first fitted item
			if (idsListItr != idsList.end())
			{
				itr->items.push_back(*idsListItr);
				itr->freeSpace -= normalWidth(*idsListItr);	
				itr->height = max(itr->height, normalHeight(*idsListItr));
				idsList.erase(idsListItr);
			}

			// second fitted item
			if (idsListItr2 != idsList.end())
			{
				itr->items.push_back(*idsListItr2);
				itr->freeSpace -= normalWidth(*idsListItr2);	
				itr->height = max(itr->height, normalHeight(*idsListItr2));
				idsList.erase(idsListItr2);
			}
		}
	}

	Cost DH::generateResult(const Levels& levels)
	{
		Point point(0, 0);

		for (const Level& level : levels)
		{
			for (Id item : level.items)
			{
				result.add(item, normalPoint(point));

				point.x += normalWidth(item);
			}
			point.x = 0;
			point.y += level.height;
		}

		return calcResultCost();
	}

	Cost DH::calc()
	{
		Levels levels;

		if (twoFit)
		{
			twoFitCalc(&levels);
		}
		else
		{
			normalCalc(&levels);
		}

		return generateResult(levels);
	}

	FFDH::LevelsIterator FFDH::chooseLevel(Levels* levels, const LevelsIterator& lastItr, Id id) const
	{
		if (!levels)
			return lastItr;

		int width = normalWidth(id);
		for (LevelsIterator itr = levels->begin(); itr != levels->end(); ++itr)
		{
			if (width <= itr->freeSpace)
			{
				return itr;
			}
		}

		return levels->end();
	}

	NFDH::LevelsIterator NFDH::chooseLevel(Levels*, const LevelsIterator& lastItr, Id) const
	{
		return lastItr;
	}

	BFDH::LevelsIterator BFDH::chooseLevel(Levels* levels, const LevelsIterator& lastItr, Id id) const
	{
		if (!levels)
			return lastItr;

		int width = normalWidth(id);
		int minResidual = maxNormalWidth();
		LevelsIterator result = levels->end();
		for (LevelsIterator itr = levels->begin(); itr != levels->end(); ++itr)
		{
			if (width <= itr->freeSpace)
			{
				int residual = itr->freeSpace - width;
				if (residual < minResidual)
				{
					minResidual = minResidual;
					result = itr;
				}
			}
		}
		return result;
	}

	string FFDH::name() const
	{
		string result;
		if (twoFit)
		{
			result = "TwoFit(" + direction.toString() + ", FFDH)";
		}
		else
		{
			result = "FFDH(" + direction.toString() + ")";
		}
		return result;
	}

	string NFDH::name() const
	{
		string result;
		if (twoFit)
		{
			result = "TwoFit(" + direction.toString() + ", NFDH)";
		}
		else
		{
			result = "NFDH(" + direction.toString() + ")";
		}
		return result;
	}

	string BFDH::name() const
	{
		string result;
		if (twoFit)
		{
			result = "TwoFit(" + direction.toString() + ", BFDH)";
		}
		else
		{
			result = "BFDH(" + direction.toString() + ")";
		}
		return result;
	}


} /* namespace SpritePack */
