/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mergemove.h"
#include "base.h"
#include "log.h"

namespace SpritePack {

	MergeMove::MergeMove()
		: compress(false), buckets(0)
	{
	}

	MergeMove::~MergeMove()
	{
	}

	Buckets::iterator MergeMove::inputItr1(const Move& move)
	{
		MergeMove * mergeMove = dynamic_cast<MergeMove*>(move.impl());
		if (mergeMove)
			return mergeMove->itr1;

		return Buckets::iterator();
	}

	Buckets::iterator MergeMove::inputItr2(const Move& move)
	{
		MergeMove * mergeMove = dynamic_cast<MergeMove*>(move.impl());
		if (mergeMove)
			return mergeMove->itr2;

		return Buckets::iterator();
	}

	Buckets::iterator MergeMove::resultItr(const Move& move)
	{
		MergeMove * mergeMove = dynamic_cast<MergeMove*>(move.impl());
		if (mergeMove)
			return mergeMove->m_resultItr;

		return Buckets::iterator();
	}

	Cost MergeMove::calc()
	{
		if (!base || !buckets || itr1 == buckets->end() || itr2 == buckets->end())
			return Cost::infinity();

		result.addItems(*itr1);
		result.addItems(*itr2);

		Cost bestCost = Cost::infinity();

		if (compress)
		{
			Bucket bestBucket;

			{
				Bucket bucket = result;
				Move move = bestMove(base->directionGeometryMoves(&bucket, Vertical));
				if (move.isGood())
				{
					move.apply();
					bestMove(base->compressionMoves(&bucket)).apply();
					Cost cost = base->cost(bucket, costType());
					if (cost < bestCost)
					{
						bestCost = cost;
						bestBucket = bucket;
					}
				}
			}

			{
				Bucket bucket = result;
				Move move = bestMove(base->directionGeometryMoves(&bucket, Horizontal));
				if (move.isGood())
				{
					move.apply();
					bestMove(base->compressionMoves(&bucket)).apply();
					Cost cost = base->cost(bucket, costType());
					if (cost < bestCost)
					{
						bestCost = cost;
						bestBucket = bucket;
					}
				}
			}

			{
				Bucket bucket = result;
				Move move = bestMove(base->directionLessGeometryMoves(&bucket));
				if (move.isGood())
				{
					move.apply();
					bestMove(base->compressionMoves(&bucket)).apply();
					Cost cost = base->cost(bucket, costType());
					if (cost < bestCost)
					{
						bestCost = cost;
						bestBucket = bucket;
					}
				}
			}

			result = bestBucket;
		}
		else
		{
			Moves moves;

			moves += bestMove(base->directionGeometryMoves(&result, Vertical));
			moves += bestMove(base->directionGeometryMoves(&result, Horizontal));
			moves += bestMove(base->directionLessGeometryMoves(&result));
			
			bestMove(moves).apply();

			bestCost = base->cost(result, costType());
		}
	
		return bestCost;
	}

	CostType MergeMove::costType() const
	{
		if (compress)
		{
			return CompressionCost;
		}
		else
		{
			return GeometryCost;
		}
	}
		
	void MergeMove::apply()
	{
		if (!buckets || !base || itr1 == buckets->end() || itr2 == buckets->end())
			return;

		const string a = "[" + toString(itr1->id()) + "|" + base->geometryMoveName(*itr1) + "|" + base->compressionMoveName(*itr1) + "|" + toString(base->cost(*itr1, costType())) + "|" + toString(base->fileSize(*itr1)) + "]";
		const string b = "[" + toString(itr2->id()) + "|" + base->geometryMoveName(*itr2) + "|" + base->compressionMoveName(*itr2) + "|" + toString(base->cost(*itr2, costType())) + "|" + toString(base->fileSize(*itr2)) + "]";
		const string c = "[" + toString(result.id()) + "|" + base->geometryMoveName(result) + "|" + base->compressionMoveName(result) + "|" + toString(base->cost(result, costType())) + "|" + toString(base->fileSize(result)) + "]";

		const Cost costBefore = base->cost(*itr1, costType()) + base->cost(*itr2, costType());
		const Cost costAfter = base->cost(result, costType());
		const string costGainPercent = costBefore != 0 ? toString(100.0 * -(costAfter - costBefore) / costBefore) + "%" : "";
		const FileSize fileSizeBefore = base->fileSize(*itr1) + base->fileSize(*itr2);
		const FileSize fileSizeAfter = base->fileSize(result);
		const string fileSizeGainPercent = fileSizeBefore != 0 ? toString(100.0 * (fileSizeAfter - fileSizeBefore) / fileSizeBefore) + "%" : "";
		const string gain = "[" + toString(-(costAfter - costBefore)) + "|" + costGainPercent + "|" + toString(fileSizeAfter - fileSizeBefore) + "|" + fileSizeGainPercent + "]";

		// Log() << "Merge:" << a + "+" + b + "=" + c + " " + gain;
			
		*itr1 = result; //replace with first
		buckets->erase(itr2); //erase second
		m_resultItr = itr1;

		if (costType() == CompressionCost || costType() == GeometryCost)
			base->geometryMoveUsed(base->geometryMoveName(result));

		if (costType() == CompressionCost)				
			base->compressionMoveUsed(base->compressionMoveName(result));
	}

	string MergeMove::name() const
	{
		return "MergeMove()";
	}

	bool MergeMove::isImprovement()
	{
		if (base && buckets)
		{
			Cost before = base->cost(*itr1, costType()) + base->cost(*itr2, costType());
			Cost after = cost();
			return after < before;
		}
		return false;
	}

	void MergeMove::setInput(bool _compress, Buckets * _buckets, const Buckets::iterator& _itr1, const Buckets::iterator& _itr2)
	{
		compress = _compress;
		buckets = _buckets;
		itr1 = _itr1;
		itr2 = _itr2;
		m_resultItr = Buckets::iterator();
	}

	MergeMoves::MergeMoves()
		: m_compress(false)
	{}

	MergeMoves& MergeMoves::operator+=(const Move& move)
	{
		Bucket::Id id1 = MergeMove::inputItr1(move)->id();
		Bucket::Id id2 = MergeMove::inputItr2(move)->id();

		m_map.insert(make_pair(id1, move));
		m_map.insert(make_pair(id2, move));

		m_moves.insert(move);

		return *this;
	}

	Move MergeMoves::bestMove() const
	{
		if (m_moves.empty() || !m_input || !m_base || m_input->size() <= 1)
			return Move();

		if (!m_compress)
			return *(m_moves.begin());

		Move bestMove;
		Cost bestCost = Cost::infinity();

		int c = (int)(m_input->size() - 1);
		double vc = m_base->speedMbps(c);
		double v1 = m_base->speedMbps(1);
		Cost startupCost = 1000 * m_base->startupMs;

		for (Set::iterator itr1 = m_moves.begin(); itr1 != m_moves.end(); ++itr1)
		{
			Move move = *itr1;
			Bucket::Id id1 = MergeMove::inputItr1(move)->id();
			Bucket::Id id2 = MergeMove::inputItr2(move)->id();
			CostType costType = CompressionCost;
			Cost mergedBucketCost = move.cost();
			Cost sumCost = 0;
			Cost maxCost = 0;

			sumCost += mergedBucketCost;
			maxCost = max(maxCost, mergedBucketCost);
			for (Buckets::iterator itr2 = m_input->begin(); itr2 != m_input->end(); ++itr2)
			{
				if (itr2->id() != id1 && itr2->id() != id2)
				{
					Cost cost = m_base->cost(*itr2, costType);
					sumCost += (double)startupCost + (cost - startupCost) * v1 / vc;
					maxCost = max(maxCost, cost);
				}
			}
			sumCost = sumCost / c;

			Cost moveTotalCost = max(sumCost, maxCost);
			if (moveTotalCost < bestCost)
			{
				bestCost = moveTotalCost;
				bestMove = move;
			}
		}

		return bestMove;
	}

	bool MergeMoves::MoveCmp::operator()(const Move& a, const Move& b) const
	{
		return a.cost() != b.cost() ? a.cost() < b.cost() : a.impl() < b.impl();
	}

	void MergeMoves::init(Base * base, Buckets * input, bool compress)
	{
		m_base = base;
		m_input = input;
		m_compress = compress;

		if (!m_base || !m_input)
			return;

		for (Buckets::iterator itr1 = m_input->begin(); itr1 != m_input->end(); ++itr1)
		{
			for (Buckets::iterator itr2 = ++Buckets::iterator(itr1); itr2 != m_input->end(); ++itr2)
			{
				*this += m_base->createMove<MergeMove>(m_compress, m_input, itr1, itr2);
			}
		}
	}

	void MergeMoves::removeDependencies(const Move& move)
	{
		Bucket::Id id1 = MergeMove::inputItr1(move)->id();
		Bucket::Id id2 = MergeMove::inputItr2(move)->id();

		for (pair<Map::iterator, Map::iterator> range = m_map.equal_range(id1); range.first != range.second; ++range.first)
		{
			const Move& move = range.first->second;
			m_moves.erase(move);
		}

		for (pair<Map::iterator, Map::iterator> range = m_map.equal_range(id2); range.first != range.second; ++range.first)
		{
			const Move& move = range.first->second;
			m_moves.erase(move);
		}

		m_map.erase(id1);

		m_map.erase(id2);
	}

	void MergeMoves::addDependencies(const Move& move)
	{
		if (!m_base || !m_input)
			return;

		Buckets::iterator itr1 = MergeMove::resultItr(move);

		for (Buckets::iterator itr2 = m_input->begin(); itr2 != m_input->end(); ++itr2)
		{
			if (itr1 != itr2)
			{
				*this += m_base->createMove<MergeMove>(m_compress, m_input, itr1, itr2);
			}
		}
	}

	bool MergeMoves::apply(const Move& move, bool improvementOnly)
	{
		if (move.isGood(improvementOnly))
		{
			removeDependencies(move);
			move.apply();	
			addDependencies(move);
			return true;
		}
		return false;
	}

	bool MergeMoves::applyIfImprovement(const Move& move)
	{
		return apply(move, true);
	}
	

} /* namespace SpritePack */
