/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "move.h"
#include "log.h"

namespace SpritePack {

	MoveImpl::MoveImpl(Base * _base)
		: base(_base), m_isCalculated(false), m_cost(Cost::infinity())
	{}

	MoveImpl::~MoveImpl()
	{
	}

	Cost MoveImpl::cost()
	{
		if (!m_isCalculated)
		{
			m_isCalculated = true;
			beforeCalc();
			m_cost = calc();			
			afterCalc();
		}
		return m_cost;
	}

	void MoveImpl::setBase(Base * _base)
	{
		base = _base;
	}

	NullMove::NullMove(Cost cost)
	{
		m_cost = cost;
	}

	Cost NullMove::calc()
	{
		return m_cost;
	}

	void NullMove::apply()
	{}

	string NullMove::name() const
	{
		return "NullMove";
	}

	bool NullMove::isImprovement()
	{
		return false;
	}

	bool Move::apply(bool improvementOnly)  const
	{
		if (isGood(improvementOnly))
		{
			m_impl->apply();
			// Log() << "move applied" << name() << cost().toString();
			return true;
		}

		return false;
	}

	bool Move::applyIfImprovement() const
	{
		return apply(true);
	}

	Cost Move::cost() const
	{
		return m_impl ? m_impl->cost() : Cost::infinity();
	}

	string Move::name() const
	{
		return m_impl ? m_impl->name() : "null";
	}

	bool Move::operator<(const Move& move) const 
	{ 
		return cost() < move.cost(); 
	}
	
	Move::Move(const RefPtr<MoveImpl>& impl)
		: m_impl(impl)
	{
		 //Log() << "move created" << name() << cost().toString();
	}

	Move::Move(Cost cost)
		: m_impl(new NullMove(cost))
	{
	}

	bool Move::isNull() const
	{
		return m_impl.isNull();
	}

	bool Move::isGood(bool improvementOnly) const
	{
		return !isNull() && !cost().isInfinity() && (!improvementOnly || (improvementOnly && m_impl->isImprovement()));
	}

	bool Move::isImprovement() const
	{
		if (m_impl)
			return m_impl->isImprovement();

		return false;
	}
	
	MoveImpl * Move::impl() const
	{
		return m_impl.get();
	}

	Move bestMove(const Moves& moves)
	{
		if (!moves.empty())
		{
			return *min_element(moves.begin(), moves.end());
		} 
		else 
		{
			return Move(Cost::infinity());
		}
	}

	Moves& operator+=(Moves& moves, const Move& move)
	{
		if (move.isGood())
			moves.push_back(move);		
		return moves;
	}

} /* namespace SpritePack */
