/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef REFPTR_H
#define REFPTR_H

namespace SpritePack
{
	class RefCounted
	{
	public:
		virtual ~RefCounted() {}

	protected:
		RefCounted() 
			: m_refCounter(0) {}

	private:
		mutable int m_refCounter;

		friend class RefPtrBase;
	};

	class RefPtrBase
	{
	protected:
		static int & refCounter(RefCounted * ptr)
		{
			return ptr->m_refCounter;
		}
	};

	template<class T> class RefPtr : private RefPtrBase
	{
	public:
		RefPtr()
			: m_ptr(0) 
		{}

		RefPtr(const RefPtr<T>& refPtr)
			: m_ptr(refPtr.m_ptr)
		{
			addRef();
		}

		RefPtr(T * t)
			: m_ptr(t)
		{
			addRef();
		}

		~RefPtr()
		{
			delRef();
		}

		RefPtr<T>& operator=(T * t)
		{
			if (t != m_ptr)
			{
				delRef();
				m_ptr = t;
				addRef();
			}
			return *this;
		}

		RefPtr<T>& operator=(const RefPtr<T>& refPtr)
		{
			if (m_ptr != refPtr.m_ptr)
			{
				delRef();
				m_ptr = refPtr.m_ptr;
				addRef();
			}
			return *this;
		}

		bool isNull() const
		{
			return m_ptr == 0;
		}

		T * get() const
		{
			return m_ptr;
		}

		T * operator->() const
		{
			return m_ptr;
		}

		operator bool() const
		{
			return m_ptr != 0;
		}

	private:
		T * m_ptr;

	private:
		void addRef()
		{
			if (m_ptr)
			{
				++refCounter(m_ptr);
			}
		}

		void delRef()
		{
			if (m_ptr)
			{	
				if (--refCounter(m_ptr) == 0)				
					delete m_ptr;				
				m_ptr = 0;
			}			
		}
	};

} /* namespace SpritePack */


#endif /* REFPTR_H */
