/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "spritepack.h"
#include "dh.h"
#include "pngcompressor.h"
#include "jpegcompressor.h"
#include "mergemove.h"
#include "log.h"
#include "palette.h"
#include "bl.h"
#include "vhtl.h"
#include "simple.h"

namespace SpritePack {

	SpritePack::SpritePack()
		: lastProgressPos(-1),
		lastProgressMax(-1),
		preprocessingFinishedTime(0),
		bucketAssignmentFinishedTime(0),
		bucketPrecompressionFinishedTime(0),
		bucketMergingFinishedTime(0),
		postprocessingFinishedTime(0)
	{}

	void SpritePack::showProgress(int pos, int max)
	{
		if (lastProgressPos == pos && lastProgressMax == max)
			return;

		lastProgressPos = pos;		
		lastProgressMax = max;

		if (max > 0) {
			Log() << toString(100 * pos / max) + "%";
		}
	}

	Moves SpritePack::compressionMoves(Bucket * bucket)
	{
		Moves moves;

		moves.push_back(createMove<JPEGCompressor>(bucket, 0));
		moves.push_back(createMove<JPEGCompressor>(bucket, MagickCore::PlaneInterlace));

		for (int filter : pngFilters())
			moves.push_back(createMove<PNGCompressor>(bucket, pngCompressionLevel, filter));

		return moves;
	}

	Moves SpritePack::directionGeometryMoves(Bucket * bucket, Direction direction)
	{
		Moves moves;
	
		if (useMBL)
			moves += createMove<BL>(bucket, direction, true);

		if (useBL)
			moves += createMove<BL>(bucket, direction, false);

		if (useTwoFit)
		{
			moves += createMove<FFDH>(bucket, direction, true);	
			moves += createMove<NFDH>(bucket, direction, true);
			moves += createMove<BFDH>(bucket, direction, true);
		}

		if (useFFDH)		
			moves += createMove<FFDH>(bucket, direction, false);		

		if (useNFDH)		
			moves += createMove<NFDH>(bucket, direction, false);		

		if (useBFDH)		
			moves += createMove<BFDH>(bucket, direction, false);
		
		return moves;
	}

	Moves SpritePack::directionLessGeometryMoves(Bucket * bucket)
	{
		Moves moves;
		if (useVHTL)
			moves += createMove<VHTL>(bucket);
		
		return moves;
	}

	struct GifPngFilter
	{
		GifPngFilter(Palette palette = Palette())
			: m_palette(palette)
		{}

		bool operator()(const ImageInfo& info)
		{
			if (m_palette == Palette())
				return info.type == Png || info.type == Gif;
			else
				return (info.type == Png || info.type == Gif) && info.palette == m_palette;
		}

	private:
		Palette m_palette;
	};

	static bool jpegFilter(const ImageInfo& info)
	{
		return info.type == Jpeg;
	}

	void SpritePack::make()
	{
		makeFull();
	}

	void SpritePack::makeOne()
	{
		/*
		Bucket bucket;
		
		bucket.setType(Jpeg);
		bucket.setQuality(jpegQuality);

		for (const ImageInfo& info : imageInfos)
			bucket.add(info, 0, 0);

		
		Id ids[] = { 20, 32, 34, 45, 49, 36, 37, 47, 53, 19, 25 };

		result.front().clear();
		for (Id id : ids)
			result.front().add(id, 0, 0);
		

		createMove<FFDH>(&bucket, Vertical).apply();

		Buckets result(1, bucket);
		saveBuckets(result);
		makePostprocess(result);
		generateHtmlCss(result);
		dumpInfo(result);
		showStats(result);
		*/
	}

	void SpritePack::dumpInfo(const Buckets&, const BeforePostProcessInfo&)
	{
		const string endl = "\r\n";

		fstream f(pathJoin(outputDir, "index.html"), fstream::app | fstream::binary | fstream::out);

		f << "<!--info" << endl;

		/*
		const string costGain = toString(100.0 * (inputCost - cost(buckets)) / inputCost) + "%";
		const string costInput = inputCost.toString(CompressionCost);
		const string costOutput = cost(buckets).toString(CompressionCost);

		const string countGain = toString(100.0 * (imageInfos.size() - buckets.size()) / imageInfos.size()) + "%";
		const string countInput = toString(imageInfos.size());
		const string countOutput = toString(buckets.size());

		const string fileSizeGain = toString(100.0 * (inputFileSize - fileSize(buckets)) / inputFileSize) + "%";
		const string fileSizeInput = toString(inputFileSize);
		const string fileSizeOutput = toString(fileSize(buckets));

		const string areaGain = toString(100.0 * (inputArea - area(buckets)) / inputArea) + "%";
		const string areaInput = toString(inputArea);
		const string areaOutput = toString(area(buckets));

		const string endl = "\r\n";

		// fstream f(pathJoin(outputDir, "info.txt"), fstream::trunc | fstream::binary | fstream::out);


		f << "inputDir: " << inputDir << endl;
		f << "outputDir: " << outputDir << endl;
		f << "jpegQuality: " << jpegQuality << endl;
		f << "pngCompressionLevel: " << pngCompressionLevel << endl;
		f << "pngFilter: " << pngFilter << endl;
		f << "k: " << k << endl;
		f << "usedK: " << usedK << endl;
		f << "classesCount: " << classesCount << endl;
		f << "startupMs: " << startupMs << endl;
		//f << "speedMbps1: " << speedMbps1 << endl;
		f << "allowLossyTypeChange: " << allowLossyTypeChange << endl;

		f << "postprocessCmd: " << postprocessCmd << endl;
		*/

		f << "loadFinishedTime: " << Time::toString(loadFinishedTime) << endl;
		f << "preprocessingFinishedTime: " << Time::toString(preprocessingFinishedTime) << endl;
		f << "bucketAssignmentFinishedTime: " << Time::toString(bucketAssignmentFinishedTime) << endl;
		f << "bucketPrecompressionFinishedTime: " << Time::toString(bucketPrecompressionFinishedTime) << endl;
		f << "bucketMergingFinishedTime: " << Time::toString(bucketMergingFinishedTime) << endl;
		f << "postprocessingFinishedTime: " << Time::toString(postprocessingFinishedTime) << endl;

		/*
		if (!postprocessCmd.empty())
		{
			const string postProcessCostGain = toString(100.0 * (beforePostProcess.first - cost(buckets)) / cost(buckets)) + "%";
			const string postProcessFileSizeGain = toString(100.0 * (beforePostProcess.second - fileSize(buckets)) / fileSize(buckets)) + "%";
			const string beforePostProcessCost = beforePostProcess.first.toString(CompressionCost);
			const string beforePostProcessFileSize = toString(beforePostProcess.second);

			f << "postProcessCostGain: " << postProcessCostGain << endl;
			f << "postProcessFileSizeGain: " << postProcessFileSizeGain << endl;
			f << "beforePostProcessCost: " << beforePostProcessCost << endl;
			f << "beforePostProcessFileSize: " << beforePostProcessFileSize << endl;
		}

		f << "costGain: " << costGain << endl;
		f << "costInput: " << costInput << endl;
		f << "costOutput: " << costOutput << endl;

		f << "countGain: " << countGain << endl;
		f << "countInput: " << countInput << endl;
		f << "countOutput: " << countOutput << endl;

		f << "fileSizeGain: " << fileSizeGain << endl;
		f << "fileSizeInput: " << fileSizeInput << endl;;
		f << "fileSizeOutput: " << fileSizeOutput << endl;

		f << "areaGain: " << areaGain << endl;
		f << "areaInput: " << areaInput << endl;
		f << "areaOutput: " << areaOutput << endl;

		f << "useBL: " << useBL << endl;
		f << "useMBL: " << useMBL << endl;
		f << "useVHTL: " << useVHTL << endl;
		f << "useTwoFit: " << useTwoFit << endl;
		f << "useFFDH: " << useFFDH << endl;
		f << "useNFDH: " << useNFDH << endl;
		f << "useBFDH: " << useBFDH << endl;

		for (int i = 1; i < speedMbpsTable.size(); ++i)
			f << "speedMbps[" + toString(i) + "]: " << speedMbpsTable[i] << endl;

		for (map<string, int>::const_iterator itr = geometryMovesStats.begin(); itr != geometryMovesStats.end(); ++itr)		
			f << "geometry_" + itr->first + ": " << itr->second << endl;

		for (map<string, int>::const_iterator itr = compressionMovesStats.begin(); itr != compressionMovesStats.end(); ++itr)
			f << "compression_" + itr->first + ": " << itr->second << endl;

		int i = 0;
		for (const Bucket& bucket : buckets)
		{
			const string prefix = "bucket_" + toString(++i) + "_";
			f << prefix + "Filename: " << bucket.fileName() << endl;
			f << prefix + "Compression: " << compressionMoveName(bucket) << endl;
			f << prefix + "Geometry: " << geometryMoveName(bucket) << endl;
			f << prefix + "Area: " << area(bucket) << endl;
			f << prefix + "Cost: " << cost(bucket, CompressionCost).toString(CompressionCost) << endl;
			f << prefix + "GeometryWaste: " << geometryWaste(bucket) << endl;
			f << prefix + "FileSize: " << fileSize(bucket) << endl;
		}
		*/

		f << "-->" << endl;
	}

	void SpritePack::makeFull()
	{
		int count = 0;	
		list< pair<Buckets, MergeMoves> > classes;
		Buckets result;

		//preprocessing
		{
			for(Palette palette : palettes())
			{
				const int n = initClasses(&classes, GifPngFilter(palette));
				setClassStats(palette.toString() + (palette.hasTransparency() ? "Transparent" : ""), n);
			}

			{
				const int n = initClasses(&classes, jpegFilter);
				setClassStats("Jpeg", n); 
			}

			classesCount = (int)classes.size();
			for (pair<Buckets, MergeMoves>& p: classes)
			{
				count += (int)p.first.size();
				p.second.init(this, &p.first, false);
			}

			int numberK = 0;
			double percentK = 0;
			
			readK(&numberK, &percentK);

			if (numberK > 0)			
				usedK = numberK;
			else			
				usedK = count * percentK / 100;

			usedK = min(count, max(classesCount + 2, usedK));
		}
		preprocessingFinishedTime = Time::elapsed();
		Log() << "Progress 1/5 Preprocessing finished";
		
		//bucket assignment
		{
			int beforeBucketAssignmentCount = count;

			while (count > usedK)
			{
				showProgress(beforeBucketAssignmentCount - count, beforeBucketAssignmentCount - usedK);

				Move bestMove;
				MergeMoves * bestMergeMoves = 0;
				for (pair<Buckets, MergeMoves>& p: classes)
				{
					Move move = p.second.bestMove();
					if (move < bestMove)
					{
						bestMove = move;
						bestMergeMoves = &p.second;
					}
				}
				bool applied = false;

				if (bestMergeMoves)
					applied = bestMergeMoves->apply(bestMove);

				if (applied)
					--count;

				if (!applied)
				{
					throwError("not applied");
				}
			}
		}
		bucketAssignmentFinishedTime = Time::elapsed();
		showProgress(-1, -1);
		Log() << "Progress 2/5 Bucket Assignment finished";

		//precompression
		{
			int beforePrecompressionMax = 0;
			int procompressionPos = 0;

			for (const pair<Buckets, MergeMoves>& p : classes)
			{
				beforePrecompressionMax += (int)p.first.size();
			}

			for (const pair<Buckets, MergeMoves>& p: classes)
			{
				for (const Bucket& bucket : p.first)
				{
					showProgress(procompressionPos, beforePrecompressionMax);
					Bucket compressedBucket = bucket;
					Move move = bestMove(compressionMoves(&compressedBucket));
					move.applyIfImprovement();
					compressedBucket.setId(bucket.id());
					result += compressedBucket;

					compressionMoveUsed(compressionMoveName(compressedBucket));
					++procompressionPos;
				}
			}
			classes.clear(); //no longer needed
		}
		bucketPrecompressionFinishedTime = Time::elapsed();
		showProgress(-1, -1);
		Log() << "Progress 3/5 Bucket Precompression finished";

		//bucket merging
		{
			int beforeBucketMergingResultSize = (int)result.size();

			Buckets bestResult = result;
			Cost bestCost = cost(result);
			MergeMoves resultMergeMoves;
			resultMergeMoves.init(this, &result, true);

			while (result.size() > 1)
			{
				showProgress(beforeBucketMergingResultSize - (int)result.size(), beforeBucketMergingResultSize - 1);

				Move bestMove = resultMergeMoves.bestMove();
				bool applied = resultMergeMoves.apply(bestMove);
				Cost resultCost = cost(result);
				if (resultCost < bestCost)
				{
					bestCost = resultCost;
					bestResult = result;
				}

				if (!applied)
				{
					throwError("Not applied");
				}
			}
			result = bestResult;
		}
		bucketMergingFinishedTime = Time::elapsed();
		showProgress(-1, -1);
		Log() << "Progress 4/5 Bucket merging finished";

		BeforePostProcessInfo beforePostProcess;
		
		//postprocessing
		{
			saveBuckets(result);
			if (!postprocessCmd.empty())
			{
				beforePostProcess.first = cost(result);
				beforePostProcess.second = fileSize(result);
			}
			makePostprocess(result);		
			postprocessingFinishedTime = Time::elapsed();
			generateHtmlCss(result, beforePostProcess);
		}
		Log() << "Progress 5/5 Postprocessing/html/css finished";		

		dumpInfo(result, beforePostProcess);
		showStats(result);
	}

} /* namespace SpritePack */
