/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "utils.h"

#ifdef _WIN32
#include "dirent.h"
#else
#include <sys/types.h>
#include <dirent.h>
#endif

namespace SpritePack {

string toLower(const string& str)
{
	string result = str;
	transform(str.begin(), str.end(), result.begin(), ::tolower);
	return result;
}

string toUpper(const string& str)
{
	string result = str;
	transform(str.begin(), str.end(), result.begin(), ::toupper);
	return result;
}

list<string> readdir(const string& directory, bool recursive)
{
	list<string> result;
	DIR *dir;
	struct dirent *ent;
	dir = opendir (directory.c_str());
	rewinddir(dir);
	if (dir != NULL)
	{
		while ((ent = readdir (dir)) != NULL)
		{
			string name = ent->d_name;

			if (name != "." && name != "..")
			{
				if (recursive && ent->d_type == DT_DIR)
				{
					list<string> subdir = readdir(pathJoin(directory, name), true);
					for (const string& s : subdir)
					{
						result.push_back(pathJoin(name, s));
					}
				}

				if (ent->d_type == DT_REG)
				{
					result.push_back(name);
				}
			}
		}
		closedir(dir);
	}
	return result;
}


string pathJoin(const string& a, const string& b)
{
	return pathNormalize(a + "/" + b);
}

string pathNormalize(const string& s)
{
	string result = s;
	
	for (char & c : result)
	{
		if (c == '\\')
		{
			c = '/';
		}
	}

	size_t pos;
	while ((pos = result.find("//")) != std::string::npos)
	{
		result.replace(pos, 2, "/");
	}

	return result;
}

string trim(const string& str, const string& chars)
{
	string result = str;
	size_t leftPos = result.find_first_not_of(chars);

	if (leftPos != string::npos)
		result.erase(0, leftPos);

	size_t rightPos = result.find_last_not_of(chars);

	if (rightPos != string::npos)
		result.erase(rightPos + 1);

	return result;
}

string ltrim(const string& str, const string& chars)
{
	string result = str;
	size_t leftPos = result.find_first_not_of(chars);

	if (leftPos != string::npos)
		result.erase(0, leftPos);

	return result;
}

string rtrim(const string& str, const string& chars)
{
	string result = str;
	size_t rightPos = result.find_last_not_of(chars);

	if (rightPos != string::npos)
		result.erase(rightPos + 1);

	return result;
}

string replaceAll(const string& str, const string& from, const string& to, bool aggressive)
{
	string result = str;
    size_t startPos = 0;
    while((startPos = result.find(from, startPos)) != string::npos)
	{
		result.replace(startPos, from.length(), to);
        startPos += to.length(); 
		if (aggressive)
			startPos = 0;
    }
	return result;
}

string pad(const string& str, int size)
{
	if (str.size() < size)
	{
		return string(size - (int)str.size(), '0') + str;
	}
	else
	{
		return str;
	}
}

template<> string toString(const double& number)
{
	return toString(number, 1);
}

template<> string toString(const float& number)
{
	return toString(number, 1);
}

string toString(double number, int precision)
{
	const long long a = (long long)number;
	const long long b = (long long)(number * pow(10, precision) - floor(number) * pow(10, precision));
	return toString(a) + "." + pad(toString(b), precision);
}

template<> string toString(const bool& b)
{
	return b ? "true" : "false";
}

template<> double fromString(const string& str)
{
	return atof(str.c_str());
}

template<> float fromString(const string& str)
{
	return atof(str.c_str());
}

string tmpPath()
{
	string result;

	if (result.empty())
	{
		const char * path = getenv("TEMP");
		if (path)
			result = path;
	}

	if (result.empty())
	{
		const char * path = getenv("TMP");
		if (path)
			result = path;
	}

	if (result.empty())
		result = "/tmp";

	return result;
}

string createTmpFilePath()
{
	string path;
	bool exists = true;
	do 
	{
		int len = 10;
		string name = "SpritePack_";
		while (len--)
			name += 'a' + (rand() % ('z' - 'a' + 1));

		path = pathJoin(tmpPath(), name);
		ifstream f(path);
		exists = f.is_open();
	} while (exists);
	return path;
}

string fileExt(const string& path)
{
	string s = trim(path);

	string::size_type pos = s.find_last_of('.');
		
	if (pos != string::npos)
		return s.substr(pos + 1);

	return s;
}

void removeFile(const string& path)
{
	remove(path.c_str());
}

vector<string>& split(const string &s, char delim, vector<string> &elems)
{
	stringstream ss(s);
	string item;
	while(getline(ss, item, delim)) 
		elems.push_back(item);
	return elems;
}


vector<string> split(const string &s, char delim)
{
	vector<string> elems;
	return split(s, delim, elems);
}

int readIntValue(const string& data, const string& match)
{
	string str = readStringValue(data, match);
	string acc;
	if (!str.empty())
	{
		bool wasNumber = false;
		for (int i = 0; i < (int)str.size(); ++i)
		{
			const char c = str[i];
			const bool isNumber = '0' <= c && c <= '9';

			if (isNumber)
				acc += c;

			if (wasNumber && !isNumber)
				break;

			wasNumber = isNumber;
		}
	}

	if (!acc.empty())
		return fromString<int>(acc);
	else
		return 0;
}

string readStringValue(const string& data, const string& match)
{
	size_t pos1 = data.find(match);
	size_t pos2 = string::npos;

	if (pos1 != string::npos)
	{
		pos2 = data.find_first_of("\n\r", pos1);
		if (pos2 != string::npos)			
			return trim(data.substr(pos1, pos2 - pos1));
	}

	return "";
}

namespace Time
{
	string toString(double elapsed)
	{
		const int seconds = (int)floor(elapsed);
		const int minutes = seconds / 60;
		const string m = SpritePack::toString(minutes);
		const string s = SpritePack::toString(seconds - 60 * minutes);
		const string ms = SpritePack::toString((int)(floor(elapsed * 1000) - 1000 * floor(elapsed)));

		return pad(m, 3) + ":" + pad(s, 2) + ":" + pad(ms, 3);
	}

	string toString()
	{
		return toString(elapsed());
	}

	double elapsed(clock_t now, clock_t startTime)
	{
		return ( now - startTime ) / (double) CLOCKS_PER_SEC;
	}

	double elapsed(clock_t now)
	{
		return elapsed(now, 0);
	}

	double elapsed()
	{
		return elapsed(now());
	}

	clock_t now()
	{
		return clock();
	}

} /* namespace Time */

} /* namespace SpritePack */
