/**
 * SpritePack
 * Copyright (C) 2015 SpritePack creators
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "vhtl.h"
#include "log.h"
#include "base.h"

namespace SpritePack {

	Area VHTL::S::calcFreeArea(const VHTL::SPoint& spoint) const
	{
		Area result = 0;
		for (int i = spoint.arrayPosition.y; i < (int)rows.size(); ++i)
		{
			int freeRowWidth = 0;
			for (int j = spoint.arrayPosition.x; j < (int)columns.size(); ++j)
			{
				const bool used = mask[i][j];

				if (!used)				
					freeRowWidth += columns[j];
				else
					break;
			}
			if (freeRowWidth > 0)
				result += rows[i] * freeRowWidth;
			else
				break;
		}
		return result;
	}

	void VHTL::S::recreateM()
	{
		const Area minArea = 1;

		m.clear();
		SPoint spoint;
		for (int j = 0; j < (int)rows.size(); ++j)
		{
			spoint.arrayPosition.y = j;
			for (int i = 0; i < (int)columns.size(); ++i)
			{
				spoint.arrayPosition.x = i;

				const Area area = calcFreeArea(spoint);

				if (area >= minArea)				
					m.insert(make_pair(area, spoint));		

				spoint.realPosition.x += columns[i];
			}
			spoint.realPosition.x = 0;
			spoint.realPosition.y += rows[j];
		}
	}

	Geometry VHTL::S::unfittedGeometry(const SPoint& spoint, Id id) const
	{
		if (!base || !isValid(spoint) || !base->isValid(id))
			return Geometry(numeric_limits<int>::max(), numeric_limits<int>::max());

		Geometry geometry = base->geometry(id);
		int maxDw = numeric_limits<int>::min();
		int dh = geometry.height;
		for (int i = spoint.arrayPosition.y; i < (int)rows.size(); ++i)
		{
			int dw = geometry.width;
			bool foundUsed = false;
			for (int j = spoint.arrayPosition.x; j < (int)columns.size(); ++j)
			{
				const bool used = mask[i][j];

				if (used)
				{
					foundUsed = true;
					break;
				}

				dw -= columns[j];

				if (dw <= 0)
					break;
			}
			maxDw = max(maxDw, dw);

			if (foundUsed)
				break;

			dh -= rows[i];

			if (dh <= 0)
				break;
		}

		return Geometry(maxDw, dh);
	}

	void VHTL::S::insert(const SPoint& spoint, Id id) 
	{
		Geometry unfitted = unfittedGeometry(spoint, id);
		if (unfitted.width <= 0 && unfitted.height <= 0)
		{
			Geometry geometry = base->geometry(id);
			int x1 = spoint.arrayPosition.x;
			int y1 = spoint.arrayPosition.y;
			int x2 = (int)columns.size();
			int y2 = (int)rows.size();

			//x2
			{
				int acc = geometry.width;
				for (int j = x1; j < (int)columns.size(); ++j)
				{
					acc -= columns[j];
					if (acc <= 0)
					{
						x2 = j + 1;
						break;
					}
				}
			}

			//y2
			{
				int acc = geometry.height;
				for (int i = y1; i < (int)rows.size(); ++i)
				{
					acc -= rows[i];
					if (acc <= 0)
					{
						y2 = i + 1;
						break;
					}
				}
			}

			//create column
			if (unfitted.width < 0)
			{
				const int dw = -unfitted.width;
				columns[x2 - 1] -= dw;
				columns.insert(columns.begin() + x2, dw);
				for (int i = 0; i < (int)rows.size(); ++i)
					mask[i].insert(mask[i].begin() + x2, mask[i][x2 - 1]);
			}
			
			//create row
			if (unfitted.height < 0)
			{
				const int dh = -unfitted.height;
				rows[y2 - 1] -= dh;
				rows.insert(rows.begin() + y2, dh);
				mask.insert(mask.begin() + y2, mask[y2 - 1]);
			}


			for (int i = y1; i < y2; ++i)
				for (int j = x1; j < x2; ++j)
					mask[i][j] = true;

			items.push_back(Bucket::Item(id, spoint.realPosition));
		}
	}

	VHTL::ArrangeItemsResult VHTL::arrangeItems(const vector<Id>& ids, const Geometry& geometry) const
	{
		if (!base)
			ArrangeItemsResult();

		S s(base, geometry);

		Geometry areaGeometry(0, 0);		

		for (Id id : ids)
		{
			const Geometry itemGeometry = base->geometry(id);
			SPoint bestSPoint(Point(numeric_limits<int>::max(), numeric_limits<int>::max()), Point(geometry.width + 1, geometry.height + 1));
			Geometry minUnfittedGeometry(itemGeometry);

			SPoint spoint;
			for (int j = 0; j < (int)s.rows.size(); ++j)
			{
				spoint.arrayPosition.y = j;
				for (int i = 0; i < (int)s.columns.size(); ++i)
				{
					spoint.arrayPosition.x = i;
					if (!s.mask[j][i] && isBetterPoint(spoint, bestSPoint, areaGeometry, itemGeometry))
					{
						Geometry unfittedGeometry = s.unfittedGeometry(spoint, id);
						if (unfittedGeometry.width <= 0 && unfittedGeometry.height <= 0)
							bestSPoint = spoint;
						minUnfittedGeometry.width = min(minUnfittedGeometry.width, unfittedGeometry.width);
						minUnfittedGeometry.height = min(minUnfittedGeometry.height, unfittedGeometry.height);
					}
					spoint.realPosition.x += s.columns[i];
				}
				spoint.realPosition.x = 0;
				spoint.realPosition.y += s.rows[j];
			}

			/*
			// optimalization code
			const Area area = itemGeometry.width * itemGeometry.height;
			s.recreateM();
			for (set< pair<Area, SPoint>, S::Cmp>::iterator itr = s.m.begin(); itr != s.m.end(); ++itr)
			{
				if (area <= itr->first)
				{
					if (isBetterPoint(itr->second, bestSPoint, areaGeometry, itemGeometry))
					{
						Geometry unfittedGeometry = s.unfittedGeometry(itr->second, id);
						if (unfittedGeometry.width <= 0 && unfittedGeometry.height <= 0)
							bestSPoint = itr->second;
						minUnfittedGeometry.width = min(minUnfittedGeometry.width, unfittedGeometry.width);
						minUnfittedGeometry.height = min(minUnfittedGeometry.height, unfittedGeometry.height);
					}
				}
				else
					break;
			}
			*/
			
			if (s.isValid(bestSPoint))
			{
				s.insert(bestSPoint, id);

				areaGeometry.width = max(areaGeometry.width, bestSPoint.realPosition.x + itemGeometry.width);
				areaGeometry.height = max(areaGeometry.height, bestSPoint.realPosition.y + itemGeometry.height);
			}
			else
				return ArrangeItemsResult(s.items, minUnfittedGeometry, id);
		}

		return ArrangeItemsResult(s.items);
	}

	Cost VHTL::calc()
	{
		if (!input || !base)
			return Cost::infinity();

		vector<Id> ids = getImageIds();
		sort(ids.begin(), ids.end(), DecreasingHeightCmp(base));

		const int maxWidth = getMaxWidth(ids);
		const int maxHeight = getMaxHeight(ids);
		const int widthSum = getWidthSum(ids);
		const int heightSum = getHeightSum(ids);

		const Area areasSum = getAreasSum(ids);

		Bucket::Items bestItems;
		Cost bestItemsCost = Cost::infinity();
		bool lastWasNonFitted = false;
		Geometry bestGeometry(widthSum, heightSum);

		Geometry geometry(widthSum, maxHeight);
		while (geometry.width >= maxWidth)
		{
			ArrangeItemsResult tmp = arrangeItems(ids, geometry);

			if (tmp.hasUnfitted())
			{
				if (lastWasNonFitted)
					geometry.height += base->imageInfo(tmp.unfittedId).height;
				else
					geometry.height += tmp.minUnfittedGeometry.height;
				lastWasNonFitted = true;
			}
			else
			{
				geometry = base->geometry(tmp.items);
				const Cost cost = base->geometryCost(tmp.items);
				if (cost < bestItemsCost)
				{
					bestItemsCost = cost;
					bestItems = tmp.items;
					bestGeometry = geometry;
				}

				do
				{
					geometry.width = geometry.width - 1;

					if (geometry.width * geometry.height < areasSum && geometry.width > 0)
					{
						geometry.height = areasSum / geometry.width;
					}
				}
				while (geometry.width * geometry.height >= bestGeometry.width * bestGeometry.height && geometry.width >= maxWidth);

				lastWasNonFitted = false;
			}
		}

		result.setItems(bestItems);

		checkForCollisions();
		return calcResultCost();
	}

	string VHTL::name() const
	{
		return "VHTL()";
	}

	void VHTL::S::visualize() const
	{
		int w = 0;
		int h = 0;

		for (int i = 0; i < (int)rows.size(); ++i)
			h += rows[i];

		for (int j = 0; j < (int)columns.size(); ++j)
			w += columns[j];

		Image img = Image(Magick::Geometry(w, h), "white");

		Point cursor(0, 0);
		for (const Bucket::Item& item : items)
		{
			if (base->isValid(item))
			{
				img.composite(base->image(item), item.x, item.y, MagickCore::CopyCompositeOp);
			}
		}

		img.strokeWidth(0);
		img.strokeColor("Transparent");
		using Magick::Quantum;
		img.fillColor(Color(QuantumRange - 1, 0, 0, QuantumRange / 2));
		cursor = Point(0, 0);
		for (int i = 0; i < (int)rows.size(); ++i)
		{
			int h = rows[i];
			for (int j = 0; j < (int)columns.size(); ++j)
			{
				int w = columns[j];
				if (mask[i][j])
				{
					img.compose(Magick::BlendCompositeOp);
					img.draw(Magick::DrawableRectangle(cursor.x, cursor.y, cursor.x + w, cursor.y + h));
				}
				cursor.x += w;
			}
			cursor.x = 0;
			cursor.y += h;
		}

		img.strokeWidth(1);
		img.strokeColor("Black");
		img.fillColor("Transparent");
		cursor = Point(0, 0);
		for (int i = 0; i < (int)rows.size(); ++i)
		{
			img.draw(Magick::DrawableLine(0, cursor.y, w, cursor.y));
			cursor.y += rows[i];
		}
		cursor = Point(0, 0);
		for (int j = 0; j < (int)columns.size(); ++j)
		{
			img.draw(Magick::DrawableLine(cursor.x, 0, cursor.x, h));
			cursor.x += columns[j];
		}
		img.write("visualize.jpg");
	}

	VHTL::SPoint::SPoint(const Point& _arrayPosition, const Point& _realPosition)
		: arrayPosition(_arrayPosition), realPosition(_realPosition)
	{
	}

	VHTL::S::S(Base * _base, const Geometry& _geometry)
		: base(_base)
	{
		columns.push_back(_geometry.width);
		rows.push_back(_geometry.height);
		mask.push_back(vector<bool>(1, false));
	}

	bool VHTL::S::isValid(const SPoint& spoint) const
	{
		return spoint.arrayPosition.x >= 0 && spoint.arrayPosition.x < (int)columns.size() && spoint.arrayPosition.y >= 0 && spoint.arrayPosition.y < (int)rows.size();
	}

	VHTL::ArrangeItemsResult::ArrangeItemsResult(const Bucket::Items& _items, const Geometry& _minUnfittedGeometry, Id _unfittedId)
		: items(_items), minUnfittedGeometry(_minUnfittedGeometry), unfittedId(_unfittedId)
	{}

	bool VHTL::ArrangeItemsResult::hasUnfitted() const
	{
		return unfittedId >= 0;
	}

	/*
	bool VHTL::isBetterPoint(const SPoint& candidate, const SPoint& currentBest, const Geometry& areaGeometry, const Geometry& itemGeometry) const
	{
		const Geometry candidateGeometry(max(areaGeometry.width, candidate.realPosition.x + itemGeometry.width), max(areaGeometry.height, candidate.realPosition.y + itemGeometry.height));
		const Geometry currentBestGeometry(max(areaGeometry.width, currentBest.realPosition.x + itemGeometry.width), max(areaGeometry.height, currentBest.realPosition.y + itemGeometry.height));

		return candidateGeometry.width * candidateGeometry.height < currentBestGeometry.width * currentBestGeometry.height;
	}
	*/

	bool VHTL::isBetterPoint(const SPoint& candidate, const SPoint& currentBest, const Geometry&, const Geometry&) const
	{
		if (candidate.arrayPosition.x != currentBest.arrayPosition.x)
		{
			return candidate.arrayPosition.x < currentBest.arrayPosition.x;
		}
		else
		{
			return candidate.arrayPosition.y < currentBest.arrayPosition.y;
		}
	}

	bool VHTL::DecreasingHeightCmp::operator()(Id a, Id b) const
	{
		if (!base)
			return false;

		if (base->imageInfo(a).width != base->imageInfo(b).width)
			return base->imageInfo(a).width > base->imageInfo(b).width;
		else
			return base->imageInfo(a).fileName < base->imageInfo(b).fileName;
	}

	int VHTL::getHeightSum(const vector<Id>& ids) const
	{
		if (!base)
			return 0;

		int result = 0;
		for (Id id: ids)
			result += base->imageInfo(id).height;

		return result;
	}

	int VHTL::getWidthSum(const vector<Id>& ids) const
	{
		if (!base)
			return 0;

		int result = 0;
		for (Id id: ids)
			result += base->imageInfo(id).width;

		return result;
	}

	int VHTL::getMaxHeight(const vector<Id>& ids) const
	{
		if (!base)
			return 0;

		int result = 0;
		for (Id id : ids)
			result = max(result, base->imageInfo(id).height);

		return result;
	}

	int VHTL::getMaxWidth(const vector<Id>& ids) const
	{
		if (!base)
			return 0;

		int result = 0;
		for (Id id : ids)
			result = max(result, base->imageInfo(id).width);

		return result;
	}

	Area VHTL::getAreasSum(const vector<Id>& ids) const
	{
		if (!base)
			return 0;

		Area result = 0;
		for (Id id : ids)
			result += (Area)base->imageInfo(id).width * (Area)base->imageInfo(id).height;

		return result;
	}

	VHTL::DecreasingHeightCmp::DecreasingHeightCmp(Base * _base)
		: base(_base)
	{}


} // namespace SpritePack
